from typing import TYPE_CHECKING

from django.db import migrations

from ipfabric_netbox.utilities.endpoint import do_endpoint_change
from ipfabric_netbox.utilities.endpoint import EndpointRecord
from ipfabric_netbox.utilities.transform_map import do_change
from ipfabric_netbox.utilities.transform_map import FieldRecord
from ipfabric_netbox.utilities.transform_map import RelationshipRecord
from ipfabric_netbox.utilities.transform_map import TransformMapRecord

if TYPE_CHECKING:
    from django.apps import apps as apps_type
    from django.db.backends.base.schema import BaseDatabaseSchemaEditor

ENDPOINTS = (
    EndpointRecord(
        name="Default VSS Chassis Endpoint",
        description="",
        endpoint="/technology/platforms/vss/chassis",
    ),
)

TRANSFORM_MAP_CHANGES = (
    TransformMapRecord(
        source_endpoint="/technology/platforms/stack/members",
        target_model="dcim.virtualchassis",
        fields=(
            FieldRecord(
                source_field="master",
                target_field="name",
                old_template="",
                new_template="{% if object.master is defined and object.master %}{{ object.master }}{% else %}{{ object.hostname }}{% endif %}",
            ),
        ),
    ),
    TransformMapRecord(
        source_endpoint="/inventory/devices",
        target_model="dcim.device",
        fields=(
            FieldRecord(
                source_field="hostname",
                target_field="vc_position",
                old_template="{% if object.virtual_chassis %}{{ object.virtual_chassis.member }}{% else %}None{% endif %}",
                new_template="{% if object.virtual_chassis %}{% if object.virtual_chassis.member is defined and object.virtual_chassis.member %}{{ object.virtual_chassis.member }}{% else %}{{ object.virtual_chassis.chassisId }}{% endif %}{% else %}None{% endif %}",
            ),
        ),
        relationships=(
            RelationshipRecord(
                source_model="dcim.virtualchassis",
                target_field="virtual_chassis",
                old_template="{% if object.virtual_chassis %}{{ dcim.VirtualChassis.objects.filter(name=object.virtual_chassis.master).first().pk }}{% endif %}",
                new_template="{% if object.virtual_chassis %}{% if object.virtual_chassis.master is defined and object.virtual_chassis.master %}{{ dcim.VirtualChassis.objects.filter(name=object.virtual_chassis.master).first().pk }}{% else %}{{ dcim.VirtualChassis.objects.filter(name=object.virtual_chassis.hostname).first().pk }}{% endif %}{% endif %}",
            ),
        ),
    ),
)


def add_vss_chassis_endpoint(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """Add VSS chassis endpoint."""
    do_endpoint_change(apps, schema_editor, ENDPOINTS, forward=True)


def remove_vss_chassis_endpoint(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """Remove VSS chassis endpoint."""
    do_endpoint_change(apps, schema_editor, ENDPOINTS, forward=False)


def forward_transform_maps_change(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """Replace old template with updated version."""
    do_change(apps, schema_editor, changes=TRANSFORM_MAP_CHANGES, forward=True)


def revert_transform_maps_change(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """Revert template back to the previous exact template."""
    do_change(apps, schema_editor, changes=TRANSFORM_MAP_CHANGES, forward=False)


def add_vss_to_child_filter(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """Add VSS chassis endpoint to Default Child Devices Filter idempotently."""
    IPFabricFilter = apps.get_model("ipfabric_netbox", "IPFabricFilter")
    IPFabricEndpoint = apps.get_model("ipfabric_netbox", "IPFabricEndpoint")

    # Get the filter
    try:
        child_filter = IPFabricFilter.objects.using(schema_editor.connection.alias).get(
            name="Default Child Devices Filter"
        )
    except IPFabricFilter.DoesNotExist:
        # Filter doesn't exist, nothing to update
        return

    # Get the VSS chassis endpoint
    try:
        vss_endpoint = IPFabricEndpoint.objects.using(
            schema_editor.connection.alias
        ).get(endpoint="/technology/platforms/vss/chassis")
    except IPFabricEndpoint.DoesNotExist:
        # Endpoint doesn't exist yet, skip (will be created by add_vss_chassis_endpoint)
        return

    # Add the endpoint to the filter if not already present
    if not child_filter.endpoints.filter(pk=vss_endpoint.pk).exists():
        child_filter.endpoints.add(vss_endpoint)


def remove_vss_from_child_filter(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """Remove VSS chassis endpoint from Default Child Devices Filter."""
    IPFabricFilter = apps.get_model("ipfabric_netbox", "IPFabricFilter")
    IPFabricEndpoint = apps.get_model("ipfabric_netbox", "IPFabricEndpoint")

    # Get the filter
    try:
        child_filter = IPFabricFilter.objects.using(schema_editor.connection.alias).get(
            name="Default Child Devices Filter"
        )
    except IPFabricFilter.DoesNotExist:
        # Filter doesn't exist, nothing to update
        return

    # Get the VSS chassis endpoint
    try:
        vss_endpoint = IPFabricEndpoint.objects.using(
            schema_editor.connection.alias
        ).get(endpoint="/technology/platforms/vss/chassis")
    except IPFabricEndpoint.DoesNotExist:
        # Endpoint doesn't exist, nothing to remove
        return

    # Remove the endpoint from the filter if present
    if child_filter.endpoints.filter(pk=vss_endpoint.pk).exists():
        child_filter.endpoints.remove(vss_endpoint)


class Migration(migrations.Migration):
    dependencies = [
        ("ipfabric_netbox", "0024_finish_filters"),
    ]

    operations = [
        migrations.RunPython(
            add_vss_chassis_endpoint,
            remove_vss_chassis_endpoint,
        ),
        migrations.RunPython(
            forward_transform_maps_change,
            revert_transform_maps_change,
        ),
        migrations.RunPython(
            add_vss_to_child_filter,
            remove_vss_from_child_filter,
        ),
    ]



import json
import logging
from collections import Counter
from copy import deepcopy
from functools import cache
from importlib import metadata
from typing import Callable
from typing import TYPE_CHECKING
from typing import TypeVar

from core.exceptions import SyncError
from core.signals import clear_events
from dcim.models import Device
from dcim.models import Site
from dcim.models import VirtualChassis
from dcim.signals import assign_virtualchassis_master

try:
    # Got added in NetBox 4.4.9
    from dcim.signals import sync_cached_scope_fields
except ImportError:
    sync_cached_scope_fields = None
from django.conf import settings
from django.core.exceptions import MultipleObjectsReturned
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Model
from django.db.models import signals
from django.utils.text import slugify
from django_tables2 import Column
from extras.events import flush_events
from ipfabric import IPFClient
from jinja2.sandbox import SandboxedEnvironment
from netbox.config import get_config
from netbox.context import events_queue
from netutils.utils import jinja2_convenience_function

from ..choices import IPFabricSourceTypeChoices
from ..choices import IPFabricSyncStatusChoices
from ..exceptions import IPAddressDuplicateError
from ..exceptions import IPAddressPrimaryAssignmentError
from ..exceptions import IPAddressPrimaryRemovalError
from ..exceptions import RequiredDependencyFailedSkip
from ..exceptions import SearchError
from ..exceptions import SyncDataError


if TYPE_CHECKING:
    from ..models import IPFabricIngestion
    from ..models import IPFabricIngestionIssue
    from ipam.models import IPAddress
    from ..models import IPFabricTransformMap
    from ..models import IPFabricSync

logger = logging.getLogger("ipfabric_netbox.utilities.ipf_utils")

ModelTypeVar = TypeVar("ModelTypeVar", bound=Model)


def slugify_text(value):
    return slugify(value)


device_serial_max_length = Device._meta.get_field("serial").max_length


def serial(data: dict) -> str:
    sn_length = len(data.get("sn"))
    serial_number = data.get("sn") if sn_length < device_serial_max_length else ""
    if not serial_number:
        serial_number = data.get("id")
    return serial_number


IPF_JINJA_FILTERS = {"slugify": slugify_text, "serial": serial}


def render_jinja2(template_code, context):
    """
    Render a Jinja2 template with the provided context. Return the rendered content.
    """
    environment = SandboxedEnvironment()
    environment.filters.update(get_config().JINJA2_FILTERS)
    environment.filters.update(IPF_JINJA_FILTERS)
    environment.filters.update(jinja2_convenience_function())
    return environment.from_string(source=template_code).render(**context)


class EventsClearer:
    """
    Handles clearing of events after defined number of objects are saved to reduce memory usage.
    ChangeLoggingMiddleware causes rest_framework Fields to be held in memory for the whole
    duration of current request. This causes high memory usage during sync jobs.

    The data is already written in DB due to branching ChangeDiff. So we can trigger
    the events clearing here to free up memory.

    For performance reasons we don't want to clear the event cache after every object is saved,
    but we also don't want to wait until the end of the sync when syncing large number of objects.
    This class helps to clear events after a defined threshold is reached.

    Without it, the memory requirements are roughly 1,5GB per 10k changes.
    """

    def __init__(self, sender: object, threshold: int = 100) -> None:
        self.sender = sender
        self.threshold = threshold
        self.counter = 0

    def increment(self) -> None:
        """
        Increment the counter and clear events if threshold is reached.

        This should not be called after every instance.snapshot() but instead
        when the instance is processed. This makes sure it's changes are synced
        as a whole. Calling it after every single snapshot() causes issues.
        """
        self.counter += 1
        if self.counter >= self.threshold:
            self.clear()

    def clear(self) -> None:
        logger.debug("Clearing events to reduce memory usage.")
        # This makes sure webhooks are sent properly
        if events := list(events_queue.get().values()):
            flush_events(events)
        # And now the queue can be cleared
        clear_events.send(sender=None)
        self.counter = 0


class IPFabric(object):
    def __init__(self, parameters=None) -> None:
        if parameters:
            self.ipf = IPFClient(**parameters, unloaded=True)
        else:
            self.ipf = IPFClient(
                **settings.PLUGINS_CONFIG["ipfabric_netbox"], unloaded=True
            )
        self.ipf._client.headers[
            "user-agent"
        ] += f'; ipfabric-netbox/{metadata.version("ipfabric-netbox")}'  # noqa: E702

    def get_table_data(self, table, device):
        filter = {"sn": ["eq", device.serial]}
        split = table.split(".")

        if len(split) == 2:
            if split[1] == "serial_ports":
                table = getattr(self.ipf.technology, split[1])
            else:
                tech = getattr(self.ipf.technology, split[0])
                table = getattr(tech, split[1])
        else:
            table = getattr(self.ipf.inventory, split[0])

        columns = self.ipf.get_columns(table.endpoint)

        columns.pop(0)

        columns = [(k, Column()) for k in columns]
        data = table.all(
            filters=filter,
        )
        return data, columns


def make_hashable(obj):
    if isinstance(obj, dict):
        return tuple((k, make_hashable(v)) for k, v in obj.items())
    elif isinstance(obj, list):
        return tuple(make_hashable(x) for x in obj)
    else:
        return obj


class DataRecord:
    """Contains all data required to sync single object to NetBox."""

    def __init__(
        self,
        model_string: str,
        data: dict,
        # These values are filled later as the record is passed down the pipeline
        context: dict | None = None,
        transform_map: "IPFabricTransformMap | None" = None,
    ):
        self.model_string = model_string
        self.data = data
        self.context = context or dict()
        self.transform_map = transform_map

    def __hash__(self):
        if self._hash is None:
            try:
                self._hash = hash(
                    (
                        self.model_string,
                        # Since the dicts are already ordered, it is safe to hash them
                        # .values() are mutable, this is fixed by tuple() to get same hash every time
                        make_hashable(self.data),
                        make_hashable(self.context),
                    )
                )
            except Exception as err:
                raise Exception(f"DATA: {self.data}") from err
        return self._hash

    def __eq__(self, other):
        return isinstance(other, DataRecord) and hash(self) == hash(other)

    # Make sure data and context are sorted by keys when stored to speed up hash calculation
    # This should be safe since they do not contain nested dicts
    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = {k: v for k, v in sorted(value.items())}
        self._hash = None  # Invalidate cached hash

    @property
    def context(self):
        return self._context

    @context.setter
    def context(self, value):
        self._context = {k: v for k, v in sorted(value.items())}
        self._hash = None  # Invalidate cached hash


def order_members(members: list[dict], sn_column: str) -> dict[str, list[dict]]:
    """Order stack members to dict, where key is master serial number and values are all members.
    For stack members table we get `memberSn` column for member serial number.
    For VSS chassis table we get `chassisSn` column for member serial number.
    """
    devices = {}

    for member in members:
        # Caution: If the snapshot is created in development mode, the `sn` field is calculated from loopback IP
        # This can be spotted by checking if `sn` is different from `memberSn` for the master device
        # Plus `sn` will be IP of loopback in hex...
        master_serial = member.get("sn")
        if master_serial and member.get(sn_column):
            if master_serial in devices:
                devices[master_serial].append(member)
            else:
                devices[master_serial] = [member]

    return devices


def prepare_devices(
    devices: list[dict], members: dict[str, list[dict]]
) -> tuple[list[dict], list[dict]]:
    """
    Prepare devices list for syncing:
     - incorporating virtual chassis members
     - handle duplicate hostnames
    """

    hostnames = [d["hostname"] for d in devices]
    counter = Counter(hostnames)

    # All devices to be synced
    all_devices = []
    # All virtual chassis to be synced
    virtualchassis = []

    for device in devices[:]:
        if counter[device["hostname"]] > 1:
            device["hostname"] = f"{device['hostname']} - ({device['sn']})"
        if child_members := members.get(device.get("sn")):
            # This device is the VC master, and we're iterating over all it's members
            for child_member in child_members:
                member_id = child_member.get("member") or child_member.get("chassisId")
                member_sn = child_member.get("memberSn") or child_member.get(
                    "chassisSn"
                )
                # There is physically no device with hostname matching the virtual chassis
                # There are only members, so "hostname/1", "hostname/2", etc.
                new_device = deepcopy(device)
                new_device["hostname"] = f"{device['hostname']}/{member_id}"
                new_device["virtual_chassis"] = child_member
                if device.get("sn") != member_sn:
                    # VC members (non-master) are not in Device table, need to add them as new Device
                    # VSS Chassis have no model in the table, get it from master device
                    new_device["model"] = child_member.get("pn") or device["model"]
                    new_device["sn"] = member_sn
                    all_devices.append(new_device)
                else:
                    # Master device, create the virtual chassis based on it
                    virtualchassis.append(child_member)
                    all_devices.append(new_device)
            hostnames = [d["hostname"] for d in devices]
            counter = Counter(hostnames)
        else:
            all_devices.append(device)
    return all_devices, virtualchassis


class IPFabricSyncRunner(object):
    def __init__(
        self,
        sync: "IPFabricSync",
        client: IPFClient = None,
        ingestion=None,
        settings: dict = None,
    ) -> None:
        self.client = client
        self.settings = settings
        self.ingestion = ingestion
        self.sync = sync
        self.transform_maps = sync.get_transform_maps(sync.parameters.get("groups"))
        if hasattr(self.sync, "logger"):
            self.logger = self.sync.logger

        if self.sync.snapshot_data.status != "loaded":
            raise SyncError("Snapshot not loaded in IP Fabric.")

        # Some objects depend on others being synced first, store errors to avoid duplicates
        # We should store all dependant objects, but it's very hard to do
        # For now store only serial numbers since that is the largest dependency chain
        self.error_serials = set()

        self.events_clearer = EventsClearer(sender=self.__class__, threshold=100)

    @staticmethod
    def get_error_serial(context: dict | None, data: dict | None) -> str | None:
        """Get error serial from context or raw data for skipping purposes."""
        context = context or {}
        data = data or {}
        return (
            context.get("sn") or context.get("serial") or serial(data)
            if "sn" in data
            else None
        )

    def create_or_get_sync_issue(
        self,
        exception: Exception,
        ingestion: "IPFabricIngestion",
        message: str = None,
        model_string: str = None,
        context: dict = None,
        data: dict = None,
    ) -> "tuple[bool, IPFabricIngestionIssue]":
        """
        Helper function to handle sync errors and create IPFabricIngestionIssue if needed.
        """
        context = context or {}

        error_serial = self.get_error_serial(context, data)
        # Ignore models that do not have any dependencies by serial number
        if error_serial and model_string not in ["ipam.ipaddress", "dcim.macaddress"]:
            self.error_serials.add(error_serial)

        # TODO: This is to prevent circular import issues, clean it up later.
        from ..models import IPFabricIngestionIssue

        if not hasattr(exception, "issue_id") or not exception.issue_id:
            issue = IPFabricIngestionIssue.objects.create(
                ingestion=ingestion,
                exception=exception.__class__.__name__,
                message=message or getattr(exception, "message", str(exception)),
                model=model_string,
                coalesce_fields={
                    k: v for k, v in context.items() if k not in ["defaults"]
                },
                defaults=context.get("defaults", dict()),
                raw_data=data or dict(),
            )
            if hasattr(exception, "issue_id"):
                exception.issue_id = issue.id
            return True, issue
        else:
            issue = IPFabricIngestionIssue.objects.get(id=exception.issue_id)
            return False, issue

    @staticmethod
    def handle_errors(func: Callable):
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as err:
                # Log the error to logger outside of job - console/file
                logger.error(err, exc_info=True)
                if hasattr(err, "issue_id") and err.issue_id:
                    # The error is already logged to user, no need to log it again
                    return None
                # Logging section for logs inside job - facing user
                self = args[0]
                if isinstance(err, SearchError):
                    if self.settings.get(err.model_string):
                        self.logger.log_failure(
                            f"Aborting syncing `{err.model_string}` instance due to above error, please check your transform maps and/or existing data.",
                            obj=self.sync,
                        )
                    else:
                        self.logger.log_failure(
                            f"Syncing `{err.model_string}` is disabled in settings, but hit above error trying to find the correct item. Please check your transform maps and/or existing data.",
                            obj=self.sync,
                        )
                if isinstance(err, IPAddressDuplicateError):
                    self.logger.log_warning(
                        f"IP Address `{err.data.get('address')}` already exists in `{err.model_string}` with coalesce fields: `{err.coalesce_fields}`. Please check your transform maps and/or existing data.",
                        obj=self.sync,
                    )
                else:
                    self.logger.log_failure(
                        f"Syncing failed with: `{err}`. See above error for more details.",
                        obj=self.sync,
                    )
                # Make sure the whole sync is failed when we encounter error
                self.sync.status = IPFabricSyncStatusChoices.FAILED
                return None

        return wrapper

    def get_db_connection_name(self) -> str:
        connection_name = None
        if self.ingestion:
            connection_name = self.ingestion.branch.connection_name
        return connection_name

    def get_transform_context(self, record: DataRecord) -> DataRecord:
        if not record.transform_map:
            raise SystemError(f"No transform map available for {record.model_string}")
        try:
            record.context = record.transform_map.get_context(record.data)
        except Exception as err:
            message = f"Error getting context for `{record.model_string}`."
            if isinstance(err, ObjectDoesNotExist):
                message += (
                    " Could not find related object using template in transform maps."
                )
            elif isinstance(err, MultipleObjectsReturned):
                message += " Multiple objects returned using on template in transform maps, the template is not strict enough."
            _, issue = self.create_or_get_sync_issue(
                exception=err,
                ingestion=self.ingestion,
                message=message,
                model_string=record.model_string,
                data=record.data,
            )
            raise SearchError(
                message=message,
                data=record.data,
                model_string=record.model_string,
                issue_id=issue.pk,
            ) from err

        return record

    def get_model_or_update(self, record: DataRecord) -> ModelTypeVar | None:
        # First check if there are any previous errors linked to this object
        error_serial = self.get_error_serial(record.context, record.data)
        if error_serial and error_serial in self.error_serials:
            # We want to raise it as exception so it's shown in ingestion issues but can be filtered out.
            exception = RequiredDependencyFailedSkip(
                message=f"Skipping syncing of `{record.model_string}` with serial `{error_serial}` due to previous errors.",
                model_string=record.model_string,
                context=record.context,
                data=record.data,
            )
            _, issue = self.create_or_get_sync_issue(
                exception=exception,
                ingestion=self.ingestion,
                model_string=record.model_string,
                context=record.context,
                data=record.data,
            )
            raise exception
        record = self.get_transform_context(record)
        queryset = record.transform_map.target_model.model_class().objects
        model_settings = self.settings.get(f"{record.model_string}", False)

        obj = None
        try:
            connection_name = self.get_db_connection_name()
            if model_settings:
                logger.info(f"Creating {record.model_string}")
                obj = record.transform_map.update_or_create_instance(
                    context=record.context,
                    tags=self.sync.tags.all(),
                    connection_name=connection_name,
                )
            else:
                logger.info(f"Getting {record.model_string}")
                record.context.pop("defaults", None)
                obj = queryset.using(connection_name).get(**record.context)
        except queryset.model.DoesNotExist as err:
            message = f"Instance of `{record.model_string}` not found."
            _, issue = self.create_or_get_sync_issue(
                exception=err,
                ingestion=self.ingestion,
                message=message,
                model_string=record.model_string,
                context=record.context,
                data=record.data,
            )
            raise SearchError(
                message=message,
                model_string=record.model_string,
                context=record.context,
                data=record.data,
                issue_id=issue.pk,
            ) from err
        except queryset.model.MultipleObjectsReturned as err:
            message = f"Multiple instances of `{record.model_string}` found."
            _, issue = self.create_or_get_sync_issue(
                exception=err,
                ingestion=self.ingestion,
                message=message,
                model_string=record.model_string,
                context=record.context,
                data=record.data,
            )
            raise SearchError(
                message=message,
                model_string=record.model_string,
                context=record.context,
                data=record.data,
                issue_id=issue.pk,
            ) from err
        except Exception as err:
            _, issue = self.create_or_get_sync_issue(
                exception=err,
                ingestion=self.ingestion,
                model_string=record.model_string,
                context=record.context,
                data=record.data,
            )
            raise SyncDataError(
                model_string=record.model_string,
                context=record.context,
                data=record.data,
                issue_id=issue.pk,
            ) from err

        return obj

    def collect_data(self):
        # Importing here to avoid circular import
        from ..models import IPFabricEndpoint

        try:
            self.logger.log_info(
                "Collecting information from IP Fabric",
                obj=self.sync.snapshot_data.source,
            )
            data = {}
            # TODO: Replace endpoint value in SnapshotData with Endpoint link
            if self.sync.snapshot_data.source.type == IPFabricSourceTypeChoices.REMOTE:
                # This requires data already pushed to NetBox by user, does not connect to IPF directly
                self.logger.log_info(
                    "Remote collector checking for snapshot data.", obj=self.sync
                )
                if not self.sync.snapshot_data.ipf_data.count() > 0:
                    raise SyncError(
                        "No snapshot data available. This is a remote sync. Push data to NetBox first."
                    )
                for endpoint in IPFabricEndpoint.objects.all():
                    data[endpoint.endpoint] = list(
                        self.sync.snapshot_data.ipf_data.filter(
                            type=endpoint.endpoint
                        ).values_list("data", flat=True)
                    )
            else:
                # This pulls data directly from IP Fabric instance
                # TODO: If certain columns are not enabled in IPF, sync will fail in odd ways
                #  For example missing `nameOriginal` on Interface makes it unable to sync IPs
                #  When pulling from endpoint, make sure to specify columns according to transform maps
                #  This can be pulled using logic in IPFabricTransformMap.strip_source_data()
                # TODO: Also cache it to improve performance, no need to regex every time!
                self.logger.log_info(
                    "Local collector being used for snapshot data.", obj=self.sync
                )
                enabled_models = [
                    param
                    for param in self.sync.parameters.keys()
                    if "." in param and self.sync.parameters.get(param) is True
                ]
                # Always pull devices for now since other models depend on them
                enabled_models.append("dcim.device")
                # Interfaces are required for IP Addresses assigned interface name
                if "ipam.ipaddress" in enabled_models:
                    enabled_models.append("dcim.interface")
                transform_maps = self.sync.get_transform_maps()
                for endpoint in IPFabricEndpoint.objects.all():
                    # FIXME: Dirty hack to sync VSS, remove when IN-68 is getting done
                    if endpoint.endpoint == "/technology/platforms/vss/chassis":
                        tms_for_endpoint = transform_maps.filter(
                            source_endpoint__endpoint="/technology/platforms/stack/members"
                        )
                    else:
                        tms_for_endpoint = transform_maps.filter(
                            source_endpoint=endpoint
                        )

                    if not tms_for_endpoint.exists():
                        raise SyncError(
                            f"No transform map found for endpoint `{endpoint.endpoint}`."
                        )
                    if not [
                        True
                        for tm in tms_for_endpoint
                        if ".".join(tm.target_model.natural_key()) in enabled_models
                    ]:
                        logger.debug(
                            f"Skipping endpoint `{endpoint.endpoint}` as no enabled models are using it."
                        )
                        continue
                    filters = endpoint.combine_filters(sync=self.sync)
                    logger.debug(
                        f"Collecting data from endpoint: `{endpoint.endpoint}` using filter `{json.dumps(filters)}`."
                    )
                    data[endpoint.endpoint] = self.client.fetch_all(
                        url=endpoint.endpoint,
                        snapshot_id=self.settings["snapshot_id"],
                        filters=filters,
                    )
                    self.logger.log_info(
                        f"Collected {len(data[endpoint.endpoint])} items from endpoint `{endpoint.endpoint}`.",
                        obj=self.sync.snapshot_data.source,
                    )
        except Exception as e:
            self.logger.log_failure(
                f"Error collecting data from IP Fabric: {e}", obj=self.sync
            )
            raise SyncError(f"Error collecting data from IP Fabric: {e}")
        return data

    @cache
    def get_transform_map(
        self, app: str, model: str, endpoint: str = None
    ) -> "IPFabricTransformMap":
        """Get transform map for given app and model. Cached to improve performance."""
        filter_kwargs = {
            "target_model__app_label": app,
            "target_model__model": model,
        }
        if endpoint:
            filter_kwargs["source_endpoint__endpoint"] = endpoint
        return self.transform_maps.get(**filter_kwargs)

    def create_new_data_records(
        self,
        app: str,
        model: str,
        data: list[dict] | dict[str, list[dict]],
        endpoint: str = None,
    ) -> set[DataRecord]:
        """Create DataRecord objects for given app, model and data list if enabled in settings."""
        if not self.sync.parameters.get(f"{app}.{model}"):
            return set()
        if isinstance(data, dict):
            # Data is either already list or full data where we have to choose the list
            transform_map = self.get_transform_map(
                app=app, model=model, endpoint=endpoint
            )
            data = data.get(transform_map.source_endpoint.endpoint, [])
        return set(
            self.create_new_data_record(
                app=app, model=model, data=item, endpoint=endpoint
            )
            for item in data
        )

    def create_new_data_record(
        self, app: str, model: str, data: dict, endpoint: str = None
    ) -> DataRecord:
        """Extract only relevant source data according to transform map configuration."""
        transform_map = self.get_transform_map(app=app, model=model, endpoint=endpoint)
        model_string = f"{app}.{model}"
        try:
            source_data = transform_map.strip_source_data(data)
        except KeyError as err:
            raise SyncError(
                f"Missing key column {err.args[0]} in source data when preparing data for {model_string}."
            ) from err
        return DataRecord(
            model_string=model_string, data=source_data, transform_map=transform_map
        )

    def _get_custom_preprocessor(
        self, model_string: str, endpoint: str
    ) -> Callable | None:
        """
        Get custom preprocessing function for a model+endpoint combination if one exists.

        Args:
            model_string: The model in 'app.model' format (e.g., 'dcim.virtualchassis')
            endpoint: The source endpoint (e.g., '/technology/platforms/stack/members')

        Returns:
            Callable preprocessor function or None if no custom preprocessor exists
        """
        preprocessors = {
            # Devices and related models processed together (from /inventory/devices)
            ("dcim.device", "/inventory/devices"): self._preprocess_devices,
            ("dcim.manufacturer", "/inventory/devices"): self._preprocess_devices,
            ("dcim.devicetype", "/inventory/devices"): self._preprocess_devices,
            ("dcim.platform", "/inventory/devices"): self._preprocess_devices,
            ("dcim.devicerole", "/inventory/devices"): self._preprocess_devices,
            # Virtual chassis from stack members
            (
                "dcim.virtualchassis",
                "/technology/platforms/stack/members",
            ): self._preprocess_devices,
            # Interfaces and MAC addresses processed together
            ("dcim.interface", "/inventory/interfaces"): self._preprocess_interfaces,
            ("dcim.macaddress", "/inventory/interfaces"): self._preprocess_interfaces,
            # IP addresses need interface data
            (
                "ipam.ipaddress",
                "/technology/addressing/managed-ip/ipv4",
            ): self._preprocess_ipaddresses,
        }

        return preprocessors.get((model_string, endpoint))

    def _preprocess_devices(self, data: dict, records: dict) -> None:
        """Custom preprocessing for devices - single iteration for efficiency."""
        if not (
            self.sync.parameters.get("dcim.device")
            or self.sync.parameters.get("dcim.virtualchassis")
            or self.sync.parameters.get("dcim.interface")
            or self.sync.parameters.get("dcim.macaddress")
            or self.sync.parameters.get("ipam.ipaddress")
        ):
            return

        # Prepare virtual chassis members once
        members = {}
        if self.sync.parameters.get("dcim.virtualchassis"):
            self.logger.log_info("Preparing virtual chassis members", obj=self.sync)
            members = order_members(
                data.get("/technology/platforms/stack/members", []), "memberSn"
            )
            members.update(
                order_members(
                    data.get("/technology/platforms/vss/chassis", []), "chassisSn"
                )
            )

        self.logger.log_info("Preparing devices", obj=self.sync)
        devices, virtualchassis = prepare_devices(
            data.get("/inventory/devices", []), members
        )

        device_primary_ips = {}

        # Single iteration creating records for all related models
        for device in devices:
            if self.sync.parameters.get("dcim.manufacturer"):
                records["dcim.manufacturer"].add(
                    self.create_new_data_record(
                        app="dcim", model="manufacturer", data=device
                    )
                )

            if self.sync.parameters.get("dcim.devicetype"):
                records["dcim.devicetype"].add(
                    self.create_new_data_record(
                        app="dcim", model="devicetype", data=device
                    )
                )

            if self.sync.parameters.get("dcim.platform"):
                records["dcim.platform"].add(
                    self.create_new_data_record(
                        app="dcim", model="platform", data=device
                    )
                )

            if self.sync.parameters.get("dcim.devicerole"):
                records["dcim.devicerole"].add(
                    self.create_new_data_record(
                        app="dcim", model="devicerole", data=device
                    )
                )

            if "virtual_chassis" not in device:
                device["virtual_chassis"] = None

            if self.sync.parameters.get("dcim.device"):
                records["dcim.device"].add(
                    self.create_new_data_record(app="dcim", model="device", data=device)
                )

            device_primary_ips[device.get("sn")] = device.get("loginIp")

        self._device_primary_ips = device_primary_ips
        records["dcim.virtualchassis"] = self.create_new_data_records(
            app="dcim", model="virtualchassis", data=virtualchassis
        )

    def _preprocess_interfaces(self, data: dict, records: dict) -> None:
        """Custom preprocessing for interfaces and MAC addresses."""
        if not (
            self.sync.parameters.get("dcim.interface")
            or self.sync.parameters.get("dcim.macaddress")
            or self.sync.parameters.get("ipam.ipaddress")
        ):
            return

        if not hasattr(self, "_device_primary_ips"):
            self._preprocess_devices(data=data, records=records)

        interface_key = "nameOriginal"
        try:
            int_transform_map = self.get_transform_map(app="dcim", model="interface")
            int_name_field_map = int_transform_map.field_maps.get(target_field="name")
            interface_key = int_name_field_map.source_field
        except Exception as e:
            self.logger.log_failure(
                f"Error collecting transform map info: {e}", obj=self.sync
            )
            raise SyncError(f"Error collecting source column name for interface: {e}")

        # Store human-readable interface names to use them later for IP Addresses
        readable_int_names = {}
        self.logger.log_info("Preparing Interfaces", obj=self.sync)

        for interface in data.get("/inventory/interfaces", []):
            if self.sync.parameters.get("dcim.interface"):
                interface_record = self.create_new_data_record(
                    app="dcim", model="interface", data=interface
                )
                interface_record.data["loginIp"] = self._device_primary_ips.get(
                    interface.get("sn")
                )
                records["dcim.interface"].add(interface_record)

            readable_int_names[
                f"{interface.get('sn')}_{interface.get('intName')}"
            ] = interface.get(interface_key)

            if self.sync.parameters.get("dcim.macaddress"):
                records["dcim.macaddress"].add(
                    self.create_new_data_record(
                        app="dcim", model="macaddress", data=interface
                    )
                )

        self._readable_int_names = readable_int_names

    def _preprocess_ipaddresses(self, data: dict, records: dict) -> None:
        """Custom preprocessing for IP addresses."""
        if not self.sync.parameters.get("ipam.ipaddress"):
            return

        if not hasattr(self, "_readable_int_names"):
            self._preprocess_interfaces(data=data, records=records)

        self.logger.log_info("Preparing IP Addresses", obj=self.sync)
        for ip in data.get("/technology/addressing/managed-ip/ipv4", []):
            ip["nameOriginal"] = self._readable_int_names.get(
                f"{ip.get('sn')}_{ip.get('intName')}"
            )
            if not ip["nameOriginal"]:
                continue

            ipaddress_record = self.create_new_data_record(
                app="ipam", model="ipaddress", data=ip
            )
            ipaddress_record.data["is_primary"] = ip.get(
                "sn"
            ) in self._device_primary_ips and self._device_primary_ips.get(
                ip.get("sn")
            ) == ipaddress_record.data.get(
                "ip"
            )
            records["ipam.ipaddress"].add(ipaddress_record)

    def preprocess_data(self, data: dict) -> dict[str, set[DataRecord]]:
        """
        Preprocess data using hierarchical transform map order.

        Iterates over transform maps in dependency order (parents before children),
        using custom preprocessing functions where available or direct record creation.
        Each transform map is processed exactly once, avoiding duplicate iterations.
        """
        # Get transform maps in hierarchical order (now returns transform maps directly!)
        group_ids = self.sync.parameters.get("groups", [])
        transform_maps_hierarchy = self.sync.get_model_hierarchy(group_ids=group_ids)

        # Initialize records dict for all models (extract unique model strings)
        records = {}
        seen_models = set()
        for transform_map in transform_maps_hierarchy:
            model_string = f"{transform_map.target_model.app_label}.{transform_map.target_model.model}"
            if model_string not in seen_models:
                records[model_string] = set()
                seen_models.add(model_string)

        # Track which custom preprocessors have run (to avoid duplication)
        # Use (preprocessor, endpoint) tuple to allow same preprocessor for different endpoints
        preprocessors_run = set()

        # Process each transform map in hierarchical order
        for transform_map in transform_maps_hierarchy:
            app = transform_map.target_model.app_label
            model = transform_map.target_model.model
            model_string = f"{app}.{model}"
            endpoint = transform_map.source_endpoint.endpoint

            # Skip if not enabled in sync parameters
            if not self.sync.parameters.get(model_string):
                continue

            # Check for custom preprocessor (pass endpoint for endpoint-specific preprocessing)
            preprocessor = self._get_custom_preprocessor(model_string, endpoint)

            # Create unique key combining preprocessor function and endpoint
            preprocessor_key = (preprocessor, endpoint) if preprocessor else None

            if preprocessor and preprocessor_key not in preprocessors_run:
                # Run custom preprocessor (may populate multiple models)
                self.logger.log_info(
                    f"Running custom preprocessor for `{model_string}` from `{endpoint}`",
                    obj=self.sync,
                )
                preprocessor(data, records)
                preprocessors_run.add(preprocessor_key)
            elif not preprocessor:
                # Standard processing: use the transform map directly (no need to filter!)
                self.logger.log_info(
                    f"Preparing `{model_string}` from `{endpoint}`", obj=self.sync
                )
                records[model_string].update(
                    self.create_new_data_records(
                        app=app, model=model, data=data, endpoint=endpoint
                    )
                )

        # Log statistics
        for model_string, records_set in records.items():
            if self.settings.get(model_string) and len(records_set):
                self.logger.init_statistics(model_string, len(records_set))
                self.logger.log_info(
                    f"Prepared {len(records_set)} items for `{model_string}` to be synced.",
                    obj=self.sync,
                )

        return records

    @handle_errors
    def sync_model(
        self,
        record: DataRecord,
        stats: bool = True,
    ) -> ModelTypeVar | None:
        """Sync a single item to NetBox."""
        if not record.data:
            return None

        instance = self.get_model_or_update(record)

        # Only log when we successfully synced the item and asked for it
        if stats and instance:
            self.logger.increment_statistics(model_string=record.model_string)

        return instance

    def sync_item(
        self,
        record: DataRecord,
        cf: bool = False,
        ingestion: "IPFabricIngestion" = None,
        stats: bool = True,
    ) -> ModelTypeVar | None:
        """Sync a single item to NetBox."""
        synced_object = self.sync_model(record=record, stats=stats)
        if synced_object is None:
            return None

        if cf:
            synced_object.snapshot()
            synced_object.custom_field_data[
                "ipfabric_source"
            ] = self.sync.snapshot_data.source.pk
            if ingestion:
                synced_object.custom_field_data["ipfabric_ingestion"] = ingestion.pk
            synced_object.save()

        return synced_object

    def sync_items(
        self,
        items: set[DataRecord],
        cf: bool = False,
        ingestion: "IPFabricIngestion" = None,
        stats: bool = True,
    ) -> None:
        """Sync list of items to NetBox."""
        if not items:
            return

        model_string = (lambda record: record.model_string)(next(iter(items)))
        if not self.settings.get(model_string):
            self.logger.log_info(
                f"Did not ask to sync {model_string}s, skipping.", obj=self.sync
            )
            return

        for item in items:
            self.sync_item(item, cf, ingestion, stats)
            self.events_clearer.increment()

    @handle_errors
    def sync_ipaddress(self, ip_address: DataRecord) -> "IPAddress | None":
        """Sync a single IP Address to NetBox, separated to use @handle_errors."""
        connection_name = self.get_db_connection_name()

        # First remove primary IP from the target object.
        # It cannot be done using hooks since there is no pre_clean at it fails on full_clean()
        try:
            ipv4_address = render_jinja2(
                ip_address.transform_map.field_maps.get(
                    target_field="address"
                ).template,
                {"object": ip_address.data},
            )
            other_device = (
                Device.objects.using(connection_name)
                .exclude(serial=serial(ip_address.data))
                .get(primary_ip4__address=ipv4_address)
            )
            other_device.snapshot()
            other_device.primary_ip4 = None
            other_device.save(using=connection_name)
        except Device.DoesNotExist:
            # There is no other device with this IP as primary, all good
            pass
        except Exception as err:
            # The transform maps might be changed, and we fail to resolve the template
            # Make sure this does not crash the sync and is handled gracefully
            _, issue = self.create_or_get_sync_issue(
                exception=err,
                ingestion=self.ingestion,
                message="Error removing primary IP from other device.",
                model_string=ip_address.model_string,
                data=ip_address.data,
            )
            self.events_clearer.increment()
            raise IPAddressPrimaryRemovalError(
                data=ip_address.data,
                model_string=ip_address.model_string,
                issue_id=issue.pk,
            ) from err

        ip_address_obj: "IPAddress | None" = self.sync_item(record=ip_address)
        if ip_address_obj is None or ip_address_obj.assigned_object is None:
            self.events_clearer.increment()
            return

        parent_device = ip_address_obj.assigned_object.parent_object

        # Now assign this IP as primary to the parent device, if not assigned yet or assigned to different IP
        if ip_address.data.get("is_primary") and (
            not parent_device.primary_ip4
            or parent_device.primary_ip4.pk != ip_address_obj.pk
        ):
            try:
                parent_device.snapshot()
                parent_device.primary_ip4 = ip_address_obj
                parent_device.save(using=connection_name)
            except Exception as err:
                _, issue = self.create_or_get_sync_issue(
                    exception=err,
                    ingestion=self.ingestion,
                    message="Error assigning primary IP to device.",
                    model_string=ip_address.model_string,
                    data=ip_address.data,
                )
                self.events_clearer.increment()
                raise IPAddressPrimaryAssignmentError(
                    data=ip_address.data,
                    model_string=ip_address.model_string,
                    issue_id=issue.pk,
                ) from err
        self.events_clearer.increment()

    @handle_errors
    def sync_ip_addresses(self, ip_addresses: set[DataRecord]) -> None:
        """
        We cannot assign primary IP in signals since IPAddress does not
        contain information whether it is primary or not. And it must be done
        on Device object, so cannot be done via Transform Maps yet since that
        would require another Transform Map for Device.
        So we need to do it manually here.

        Cleaning events queue happens during each cycle to make sure all required
        operations (primary IP assignment) happen during the same batch.
        """
        if not self.settings.get("ipam.ipaddress"):
            self.logger.log_info(
                "Did not ask to sync ipaddresses, skipping.", obj=self.sync
            )
            return

        for ip_address in ip_addresses:
            self.sync_ipaddress(ip_address)

    def collect_and_sync(self, ingestion=None) -> None:
        self.logger.log_info("Starting data collection.", obj=self.sync)
        data = self.collect_data()
        self.logger.log_info("Starting to prepare items.", obj=self.sync)
        records = self.preprocess_data(data=data)

        self.logger.log_info("Starting data sync.", obj=self.sync)
        try:
            # This signal does not call for snapshot(), causing issue with branching plugin
            if sync_cached_scope_fields is not None:
                signals.post_save.disconnect(sync_cached_scope_fields, sender=Site)
            self.sync_items(
                items=records["dcim.site"],
                cf=self.sync.update_custom_fields,
                ingestion=ingestion,
            )
        finally:
            if sync_cached_scope_fields is not None:
                signals.post_save.connect(sync_cached_scope_fields, sender=Site)
        self.sync_items(items=records["dcim.manufacturer"])
        self.sync_items(items=records["dcim.devicetype"])
        self.sync_items(items=records["dcim.platform"])
        self.sync_items(items=records["dcim.devicerole"])
        try:
            # This signal does not call for snapshot(), causing issue with branching plugin
            signals.post_save.disconnect(
                assign_virtualchassis_master, sender=VirtualChassis
            )
            self.sync_items(items=records["dcim.virtualchassis"])
            self.sync_items(
                items=records["dcim.device"],
                cf=self.sync.update_custom_fields,
                ingestion=ingestion,
            )
            # The Device exists now, so we can update the master of the VC.
            # The logic is handled in transform maps.
            self.sync_items(items=records["dcim.virtualchassis"], stats=False)
        finally:
            signals.post_save.connect(
                assign_virtualchassis_master, sender=VirtualChassis
            )
        self.sync_items(items=records["dcim.interface"])
        self.sync_items(items=records["dcim.macaddress"])
        self.sync_items(items=records["dcim.inventoryitem"])
        self.sync_items(items=records["ipam.vlan"])
        self.sync_items(items=records["ipam.vrf"])
        self.sync_items(items=records["ipam.prefix"])
        self.sync_ip_addresses(ip_addresses=records["ipam.ipaddress"])

        # Make sure to clean queue (and memory) at the end
        self.events_clearer.clear()



# Generated by Django 5.0.5 on 2024-05-16 15:23
import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("ipfabric_netbox", "0004_ipfabricsync_auto_merge"),
    ]

    operations = [
        migrations.AlterField(
            model_name="ipfabricrelationshipfield",
            name="source_model",
            field=models.ForeignKey(
                limit_choices_to=models.Q(
                    models.Q(
                        ("app_label", "dcim"),
                        ("app_label", "ipam"),
                        ("app_label", "tenancy"),
                        models.Q(
                            ("app_label", "contenttypes"), ("model", "contenttype")
                        ),
                        _connector="OR",
                    )
                ),
                on_delete=django.db.models.deletion.PROTECT,
                related_name="ipfabric_transform_fields",
                to="contenttypes.contenttype",
            ),
        ),
        migrations.AlterField(
            model_name="ipfabrictransformmap",
            name="target_model",
            field=models.ForeignKey(
                limit_choices_to=models.Q(
                    models.Q(
                        models.Q(("app_label", "dcim"), ("model", "site")),
                        models.Q(("app_label", "dcim"), ("model", "manufacturer")),
                        models.Q(("app_label", "dcim"), ("model", "platform")),
                        models.Q(("app_label", "dcim"), ("model", "devicerole")),
                        models.Q(("app_label", "dcim"), ("model", "devicetype")),
                        models.Q(("app_label", "dcim"), ("model", "device")),
                        models.Q(("app_label", "dcim"), ("model", "virtualchassis")),
                        models.Q(("app_label", "dcim"), ("model", "interface")),
                        models.Q(("app_label", "ipam"), ("model", "vlan")),
                        models.Q(("app_label", "ipam"), ("model", "vrf")),
                        models.Q(("app_label", "ipam"), ("model", "prefix")),
                        models.Q(("app_label", "ipam"), ("model", "ipaddress")),
                        models.Q(
                            ("app_label", "contenttypes"), ("model", "contenttype")
                        ),
                        models.Q(("app_label", "tenancy"), ("model", "tenant")),
                        _connector="OR",
                    )
                ),
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="contenttypes.contenttype",
            ),
        ),
    ]



from core.api.serializers_.jobs import JobSerializer
from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.http import HttpResponseBadRequest
from drf_spectacular.utils import extend_schema
from netbox.api.viewsets import NetBoxModelViewSet
from netbox.api.viewsets import NetBoxReadOnlyModelViewSet
from rest_framework.decorators import action
from rest_framework.response import Response
from utilities.query import count_related

from .serializers import EmptySerializer
from .serializers import IPFabricEndpointSerializer
from .serializers import IPFabricFilterExpressionSerializer
from .serializers import IPFabricFilterSerializer
from .serializers import IPFabricIngestionIssueSerializer
from .serializers import IPFabricIngestionSerializer
from .serializers import IPFabricRelationshipFieldSerializer
from .serializers import IPFabricSnapshotSerializer
from .serializers import IPFabricSourceSerializer
from .serializers import IPFabricSyncSerializer
from .serializers import IPFabricTransformFieldSerializer
from .serializers import IPFabricTransformMapGroupSerializer
from .serializers import IPFabricTransformMapSerializer
from ipfabric_netbox.filtersets import IPFabricRelationshipFieldFilterSet
from ipfabric_netbox.filtersets import IPFabricSnapshotFilterSet
from ipfabric_netbox.filtersets import IPFabricSourceFilterSet
from ipfabric_netbox.filtersets import IPFabricTransformFieldFilterSet
from ipfabric_netbox.filtersets import IPFabricTransformMapFilterSet
from ipfabric_netbox.models import IPFabricData
from ipfabric_netbox.models import IPFabricEndpoint
from ipfabric_netbox.models import IPFabricFilter
from ipfabric_netbox.models import IPFabricFilterExpression
from ipfabric_netbox.models import IPFabricIngestion
from ipfabric_netbox.models import IPFabricIngestionIssue
from ipfabric_netbox.models import IPFabricRelationshipField
from ipfabric_netbox.models import IPFabricSnapshot
from ipfabric_netbox.models import IPFabricSource
from ipfabric_netbox.models import IPFabricSync
from ipfabric_netbox.models import IPFabricTransformField
from ipfabric_netbox.models import IPFabricTransformMap
from ipfabric_netbox.models import IPFabricTransformMapGroup


class IPFabricEndpointViewSet(NetBoxReadOnlyModelViewSet):
    queryset = IPFabricEndpoint.objects.all()
    serializer_class = IPFabricEndpointSerializer


class IPFabricTransformMapGroupViewSet(NetBoxModelViewSet):
    queryset = IPFabricTransformMapGroup.objects.all()
    serializer_class = IPFabricTransformMapGroupSerializer


class IPFabricTransformMapViewSet(NetBoxModelViewSet):
    queryset = IPFabricTransformMap.objects.all()
    serializer_class = IPFabricTransformMapSerializer
    filterset_class = IPFabricTransformMapFilterSet


class IPFabricTransformFieldViewSet(NetBoxModelViewSet):
    queryset = IPFabricTransformField.objects.all()
    serializer_class = IPFabricTransformFieldSerializer
    filterset_class = IPFabricTransformFieldFilterSet


class IPFabricRelationshipFieldViewSet(NetBoxModelViewSet):
    queryset = IPFabricRelationshipField.objects.all()
    serializer_class = IPFabricRelationshipFieldSerializer
    filterset_class = IPFabricRelationshipFieldFilterSet


class IPFabricSyncViewSet(NetBoxModelViewSet):
    queryset = IPFabricSync.objects.all()
    serializer_class = IPFabricSyncSerializer

    @extend_schema(
        methods=["post"],
        request=EmptySerializer(),
        responses={201: JobSerializer()},
    )
    @action(detail=True, methods=["post"])
    def sync(self, request, pk):
        if not request.user.has_perm("ipfabric_netbox.sync_ipfabricsync"):
            raise PermissionDenied(
                "This user does not have permission to sync IPFabricSync."
            )
        sync = self.get_object()
        if not sync.ready_for_sync:
            return HttpResponseBadRequest(
                f"Sync '{sync.name}' is not ready to be synced."
            )
        job = sync.enqueue_sync_job(user=request.user, adhoc=True)
        return Response(
            JobSerializer(job, context={"request": request}).data, status=201
        )


class IPFabricFilterViewSet(NetBoxModelViewSet):
    queryset = IPFabricFilter.objects.all()
    serializer_class = IPFabricFilterSerializer


class IPFabricFilterExpressionViewSet(NetBoxModelViewSet):
    queryset = IPFabricFilterExpression.objects.all()
    serializer_class = IPFabricFilterExpressionSerializer


class IPFabricIngestionViewSet(NetBoxReadOnlyModelViewSet):
    queryset = IPFabricIngestion.objects.all()
    serializer_class = IPFabricIngestionSerializer


class IPFabricIngestionIssueViewSet(NetBoxReadOnlyModelViewSet):
    queryset = IPFabricIngestionIssue.objects.all()
    serializer_class = IPFabricIngestionIssueSerializer


class IPFabricSnapshotViewSet(NetBoxReadOnlyModelViewSet):
    queryset = IPFabricSnapshot.objects.all()
    serializer_class = IPFabricSnapshotSerializer
    filterset_class = IPFabricSnapshotFilterSet

    @action(detail=True, methods=["patch", "delete"], url_path="raw")
    def raw(self, request, pk):
        snapshot = self.get_object()
        if request.method == "DELETE":
            raw_data = IPFabricData.objects.filter(snapshot_data=snapshot)
            raw_data._raw_delete(raw_data.db)
            return Response({"status": "success"})
        elif request.method == "PATCH":
            with transaction.atomic():
                IPFabricData.objects.bulk_create(
                    [
                        IPFabricData(snapshot_data=snapshot, data=item["data"])
                        for item in request.data["data"]
                    ],
                    batch_size=5000,
                )
            return Response({"status": "success"})

    @action(detail=True, methods=["get"], url_path="sites")
    def sites(self, request, pk):
        q = request.GET.get("q", None)
        snapshot = IPFabricSnapshot.objects.get(pk=pk)
        new_sites = {"count": 0, "results": []}
        if snapshot.data:
            sites = snapshot.data.get("sites", None)
            num = 0
            if sites:
                for site in sites:
                    if q:
                        if q.lower() in site.lower():
                            new_sites["results"].append(
                                {"display": site, "name": site, "id": site}
                            )
                    else:
                        new_sites["results"].append(
                            {"display": site, "name": site, "id": site}
                        )
                    num += 1
                new_sites["count"] = num
                return Response(new_sites)
        else:
            return Response([])


class IPFabricSourceViewSet(NetBoxModelViewSet):
    queryset = IPFabricSource.objects.annotate(
        snapshot_count=count_related(IPFabricSnapshot, "source")
    )
    serializer_class = IPFabricSourceSerializer
    filterset_class = IPFabricSourceFilterSet

    @extend_schema(
        methods=["post"],
        request=EmptySerializer(),
        responses={201: JobSerializer()},
    )
    @action(detail=True, methods=["post"])
    def sync(self, request, pk):
        if not request.user.has_perm("ipfabric_netbox.sync_ipfabricsource"):
            raise PermissionDenied(
                "This user does not have permission to sync IPFabricSource."
            )
        source = self.get_object()
        if not source.ready_for_sync:
            return HttpResponseBadRequest(
                f"Source '{source.name}' is not ready to be synced."
            )
        job = source.enqueue_sync_job(request=request)
        return Response(
            JobSerializer(job, context={"request": request}).data, status=201
        )



import json

from core.choices import ObjectChangeActionChoices
from dcim.models import Device
from dcim.models import Site
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.db import models
from django.db import transaction
from django.http import HttpResponse
from django.shortcuts import get_object_or_404
from django.shortcuts import redirect
from django.shortcuts import render
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext as _
from django.views.generic import View
from django_tables2 import RequestConfig
from ipfabric.diagrams import Network
from ipfabric.diagrams import NetworkSettings
from netbox.object_actions import AddObject
from netbox.object_actions import BulkDelete
from netbox.object_actions import BulkEdit
from netbox.object_actions import BulkExport
from netbox.object_actions import BulkRename
from netbox.views import generic
from netbox.views.generic.base import BaseObjectView
from netbox_branching.models import ChangeDiff
from utilities.data import shallow_compare_dict
from utilities.forms import ConfirmationForm
from utilities.forms import restrict_form_fields
from utilities.paginator import EnhancedPaginator
from utilities.paginator import get_paginate_count
from utilities.query import count_related
from utilities.views import get_viewname
from utilities.views import GetRelatedModelsMixin
from utilities.views import register_model_view
from utilities.views import ViewTab

from .choices import IPFabricSourceTypeChoices
from .filtersets import IPFabricDataFilterSet
from .filtersets import IPFabricEndpointFilterSet
from .filtersets import IPFabricFilterExpressionFilterSet
from .filtersets import IPFabricFilterFilterSet
from .filtersets import IPFabricIngestionChangeFilterSet
from .filtersets import IPFabricIngestionFilterSet
from .filtersets import IPFabricIngestionIssueFilterSet
from .filtersets import IPFabricSnapshotFilterSet
from .filtersets import IPFabricSourceFilterSet
from .filtersets import IPFabricSyncFilterSet
from .filtersets import IPFabricTransformMapFilterSet
from .filtersets import IPFabricTransformMapGroupFilterSet
from .forms import IPFabricFilterBulkEditForm
from .forms import IPFabricFilterBulkImportForm
from .forms import IPFabricFilterExpressionBulkEditForm
from .forms import IPFabricFilterExpressionBulkImportForm
from .forms import IPFabricFilterExpressionForm
from .forms import IPFabricFilterForm
from .forms import IPFabricIngestionFilterForm
from .forms import IPFabricIngestionMergeForm
from .forms import IPFabricRelationshipFieldForm
from .forms import IPFabricSnapshotFilterForm
from .forms import IPFabricSourceBulkEditForm
from .forms import IPFabricSourceFilterForm
from .forms import IPFabricSourceForm
from .forms import IPFabricSyncBulkEditForm
from .forms import IPFabricSyncForm
from .forms import IPFabricTableForm
from .forms import IPFabricTransformFieldForm
from .forms import IPFabricTransformMapBulkEditForm
from .forms import IPFabricTransformMapBulkImportForm
from .forms import IPFabricTransformMapCloneForm
from .forms import IPFabricTransformMapForm
from .forms import IPFabricTransformMapGroupBulkEditForm
from .forms import IPFabricTransformMapGroupBulkImportForm
from .forms import IPFabricTransformMapGroupForm
from .models import IPFabricData
from .models import IPFabricEndpoint
from .models import IPFabricFilter
from .models import IPFabricFilterExpression
from .models import IPFabricIngestion
from .models import IPFabricIngestionIssue
from .models import IPFabricRelationshipField
from .models import IPFabricSnapshot
from .models import IPFabricSource
from .models import IPFabricSync
from .models import IPFabricTransformField
from .models import IPFabricTransformMap
from .models import IPFabricTransformMapGroup
from .tables import DeviceIPFTable
from .tables import IPFabricDataTable
from .tables import IPFabricEndpointTable
from .tables import IPFabricFilterExpressionTable
from .tables import IPFabricFilterTable
from .tables import IPFabricIngestionChangesTable
from .tables import IPFabricIngestionIssuesTable
from .tables import IPFabricIngestionTable
from .tables import IPFabricRelationshipFieldTable
from .tables import IPFabricSnapshotTable
from .tables import IPFabricSourceTable
from .tables import IPFabricSyncTable
from .tables import IPFabricTransformFieldTable
from .tables import IPFabricTransformMapGroupTable
from .tables import IPFabricTransformMapTable
from .utilities.filters import get_filter_expression_test_candidates
from .utilities.ipfutils import IPFabric
from .utilities.transform_map import build_transform_maps
from .utilities.transform_map import get_transform_map


# region - Base Classes for Combined Expressions Views


class CombinedExpressionsBaseView(LoginRequiredMixin, View):
    """Base view for displaying combined filter expressions in an HTMX modal.

    This base class provides common functionality for views that display
    combined/merged filter expressions. Subclasses must implement
    get_context_data() to provide view-specific context.

    The template (inc/combined_expressions.html) supports three context types:
    - "filter": Display merged expressions for a single IPFabricFilter
    - "endpoint": Display combined filters for an endpoint within a sync
    - "endpoint_all": Display combined filters for an endpoint across all syncs
    """

    template_name = "ipfabric_netbox/inc/combined_expressions.html"

    def get_context_data(self, request, **kwargs):
        """Override this method to provide view-specific context.

        Must return a dict with standardized keys:
        - object: The main object being displayed (IPFabricFilter or IPFabricEndpoint, or None)
        - combined_expressions: The expressions (list for filters, dict for endpoints)
        - context_type: One of "filter", "endpoint", or "endpoint_all"
        - is_empty: Boolean indicating if there are no expressions
        - sync: Optional IPFabricSync object (for endpoint context type only)

        Args:
            request: The HTTP request object
            **kwargs: URL parameters passed to the view

        Returns:
            dict: Context dictionary for template rendering
        """
        raise NotImplementedError(
            "Subclasses must implement get_context_data()"
        )  # pragma: no cover

    def add_no_cache_headers(self, response):
        """Add cache control headers to prevent browser caching.

        Args:
            response: HttpResponse object

        Returns:
            HttpResponse: The response with cache headers added
        """
        response["Cache-Control"] = "no-cache, no-store, must-revalidate"
        response["Pragma"] = "no-cache"
        response["Expires"] = "0"
        return response

    def get(self, request, **kwargs):
        """Handle GET requests by rendering the template with context.

        Args:
            request: The HTTP request object
            **kwargs: URL parameters

        Returns:
            HttpResponse: Rendered template with no-cache headers
        """
        context = self.get_context_data(request, **kwargs)
        response = render(request, self.template_name, context)
        return self.add_no_cache_headers(response)


# endregion


# region - Endpoint (Read-Only)
@register_model_view(IPFabricEndpoint, "list", path="", detail=False)
class IPFabricEndpointListView(generic.ObjectListView):
    queryset = IPFabricEndpoint.objects.annotate(
        filters_count=models.Count("filters", distinct=True)
    )
    table = IPFabricEndpointTable
    filterset = IPFabricEndpointFilterSet
    template_name = "ipfabric_netbox/htmx_list.html"


@register_model_view(IPFabricEndpoint)
class IPFabricEndpointView(GetRelatedModelsMixin, generic.ObjectView):
    queryset = IPFabricEndpoint.objects.all()
    actions = ()

    def get_extra_context(self, request, instance):
        return {
            "related_models": self.get_related_models(
                request,
                instance,
                extra=(
                    (
                        IPFabricFilter.objects.restrict(request.user, "view").filter(
                            endpoints=instance
                        ),
                        "endpoints",
                    ),
                ),
            ),
        }


@register_model_view(
    IPFabricEndpoint,
    name="filters",
    path="filters",
    kwargs={"model": IPFabricEndpoint},
)
class IPFabricEndpointFiltersView(CombinedExpressionsBaseView):
    """Display combined filters for an endpoint with optional sync selection."""

    template_name = "ipfabric_netbox/inc/endpoint_filters_with_selector.html"

    def get_context_data(self, request, **kwargs):
        """Provide context for endpoint filters with sync selection.

        Args:
            request: HTTP request (must be HTMX)
            **kwargs: Must contain 'pk' (endpoint primary key)

        Returns:
            dict: Standardized context with endpoint object, available syncs,
                  and combined filters (optionally filtered by sync)
        """
        endpoint_pk = kwargs.get("pk")
        sync_pk = request.GET.get("sync_pk")
        from_sync = request.GET.get("from_sync") == "true"

        # Return empty context if not HTMX or missing required parameters
        if not request.htmx or not endpoint_pk:
            return {
                "object": None,
                "combined_expressions": {},
                "is_empty": True,
                "context_type": "endpoint_all",
            }

        # Fetch endpoint
        endpoint_obj = IPFabricEndpoint.objects.prefetch_related("filters").get(
            pk=endpoint_pk
        )

        # Determine if we're filtering by sync
        sync_obj = None
        if sync_pk:
            try:
                sync_obj = IPFabricSync.objects.get(pk=sync_pk)
            except IPFabricSync.DoesNotExist:
                pass

        # Combine filters (with optional sync filter)
        combined = endpoint_obj.combine_filters(sync=sync_obj)

        # Determine context type based on sync selection
        context_type = "endpoint" if sync_obj else "endpoint_all"

        context = {
            "object": endpoint_obj,
            "combined_expressions": combined,
            "is_empty": len(combined) == 0,
            "context_type": context_type,
            "from_sync": from_sync,
        }

        # Add sync to context if one is selected
        if sync_obj:
            context["sync"] = sync_obj

        # Only add sync selector context if not coming from a sync view
        if not from_sync:
            context["available_syncs"] = IPFabricSync.objects.all().order_by("name")
            context["selected_sync_pk"] = int(sync_pk) if sync_pk else None

        return context

    def get(self, request, **kwargs):
        """Handle GET requests, using appropriate template based on context.

        Args:
            request: The HTTP request object
            **kwargs: URL parameters

        Returns:
            HttpResponse: Rendered template with no-cache headers
        """
        context = self.get_context_data(request, **kwargs)

        # Determine which template to use
        from_sync = request.GET.get("from_sync") == "true"
        sync_pk_param = request.GET.get("sync_pk")

        if from_sync:
            # Coming from sync view - use simple template without selector
            template_name = "ipfabric_netbox/inc/combined_expressions.html"
        elif sync_pk_param is not None and not from_sync:
            # Sync selection change via dropdown - use content-only template
            template_name = "ipfabric_netbox/inc/combined_expressions_content.html"
        else:
            # Initial load from general endpoints - use template with selector
            template_name = self.template_name

        response = render(request, template_name, context)
        return self.add_no_cache_headers(response)


# endregion

# region - Transform Map Relationship Field


@register_model_view(IPFabricRelationshipField, "add", detail=False)
@register_model_view(IPFabricRelationshipField, "edit")
class IPFabricRelationshipFieldEditView(generic.ObjectEditView):
    queryset = IPFabricRelationshipField.objects.all()
    form = IPFabricRelationshipFieldForm
    default_return_url = "plugins:ipfabric_netbox:ipfabricrelationshipfield_list"


@register_model_view(IPFabricRelationshipField, "delete")
class IPFabricRelationshipFieldDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricRelationshipField.objects.all()
    default_return_url = "plugins:ipfabric_netbox:ipfabricrelationshipfield_list"


@register_model_view(
    IPFabricRelationshipField, "bulk_delete", path="delete", detail=False
)
class IPFabricRelationshipFieldBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricRelationshipField.objects.all()
    table = IPFabricRelationshipFieldTable


@register_model_view(IPFabricRelationshipField, "list", path="", detail=False)
class IPFabricRelationshipFieldListView(generic.ObjectListView):
    queryset = IPFabricRelationshipField.objects.all()
    table = IPFabricRelationshipFieldTable
    actions = ()


# endregion
# region - Transform Map Group


@register_model_view(IPFabricTransformMapGroup, "list", path="", detail=False)
class IPFabricTransformMapGroupListView(generic.ObjectListView):
    queryset = IPFabricTransformMapGroup.objects.annotate(
        maps_count=models.Count("transform_maps")
    )
    table = IPFabricTransformMapGroupTable
    filterset = IPFabricTransformMapGroupFilterSet


@register_model_view(IPFabricTransformMapGroup, "add", detail=False)
@register_model_view(IPFabricTransformMapGroup, "edit")
class IPFabricTransformMapGroupEditView(generic.ObjectEditView):
    queryset = IPFabricTransformMapGroup.objects.all()
    form = IPFabricTransformMapGroupForm
    default_return_url = "plugins:ipfabric_netbox:ipfabrictransformmapgroup_list"


@register_model_view(IPFabricTransformMapGroup, "delete")
class IPFabricTransformMapGroupDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricTransformMapGroup.objects.all()
    default_return_url = "plugins:ipfabric_netbox:ipfabrictransformmapgroup_list"


@register_model_view(
    IPFabricTransformMapGroup, "bulk_import", path="import", detail=False
)
class IPFabricTransformMapGroupBulkImportView(generic.BulkImportView):
    queryset = IPFabricTransformMapGroup.objects.all()
    model_form = IPFabricTransformMapGroupBulkImportForm


@register_model_view(IPFabricTransformMapGroup, "bulk_edit", path="edit", detail=False)
class IPFabricTransformMapGroupBulkEditView(generic.BulkEditView):
    queryset = IPFabricTransformMapGroup.objects.all()
    table = IPFabricTransformMapGroupTable
    form = IPFabricTransformMapGroupBulkEditForm


@register_model_view(
    IPFabricTransformMapGroup, "bulk_rename", path="rename", detail=False
)
class IPFabricTransformMapGroupBulkRenameView(generic.BulkRenameView):
    queryset = IPFabricTransformMapGroup.objects.all()


@register_model_view(
    IPFabricTransformMapGroup, "bulk_delete", path="delete", detail=False
)
class IPFabricTransformMapGroupBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricTransformMapGroup.objects.all()
    table = IPFabricTransformMapGroupTable


@register_model_view(IPFabricTransformMapGroup)
class IPFabricTransformMapGroupView(GetRelatedModelsMixin, generic.ObjectView):
    queryset = IPFabricTransformMapGroup.objects.all()

    def get_extra_context(self, request, instance):
        return {
            "related_models": self.get_related_models(request, instance, omit=[]),
        }


# endregion
# region - Transform Map


@register_model_view(IPFabricTransformMap, "list", path="", detail=False)
class IPFabricTransformMapListView(generic.ObjectListView):
    queryset = IPFabricTransformMap.objects.all()
    table = IPFabricTransformMapTable
    template_name = "ipfabric_netbox/ipfabrictransformmap_list.html"
    filterset = IPFabricTransformMapFilterSet


@register_model_view(IPFabricTransformMap, "add", detail=False)
@register_model_view(IPFabricTransformMap, "edit")
class IPFabricTransformMapEditView(generic.ObjectEditView):
    queryset = IPFabricTransformMap.objects.all()
    form = IPFabricTransformMapForm
    default_return_url = "plugins:ipfabric_netbox:ipfabrictransformmap_list"


@register_model_view(IPFabricTransformMap, "delete")
class IPFabricTransformMapDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricTransformMap.objects.all()
    default_return_url = "plugins:ipfabric_netbox:ipfabrictransformmap_list"


@register_model_view(IPFabricTransformMap, "bulk_import", path="import", detail=False)
class IPFabricTransformMapBulkImportView(generic.BulkImportView):
    queryset = IPFabricTransformMap.objects.all()
    model_form = IPFabricTransformMapBulkImportForm


@register_model_view(IPFabricTransformMap, "bulk_edit", path="edit", detail=False)
class IPFabricTransformMapBulkEditView(generic.BulkEditView):
    queryset = IPFabricTransformMap.objects.all()
    table = IPFabricTransformMapTable
    form = IPFabricTransformMapBulkEditForm


@register_model_view(IPFabricTransformMap, "bulk_rename", path="rename", detail=False)
class IPFabricTransformMapBulkRenameView(generic.BulkRenameView):
    queryset = IPFabricTransformMap.objects.all()


@register_model_view(IPFabricTransformMap, "bulk_delete", path="delete", detail=False)
class IPFabricTransformMapBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricTransformMap.objects.all()
    table = IPFabricTransformMapTable


@register_model_view(IPFabricTransformMap)
class IPFabricTransformMapView(GetRelatedModelsMixin, generic.ObjectView):
    queryset = IPFabricTransformMap.objects.all()

    def get_extra_context(self, request, instance):
        return {
            "related_models": self.get_related_models(
                request,
                instance,
            ),
        }


@register_model_view(IPFabricTransformMap, "restore", detail=False)
class IPFabricTransformMapRestoreView(generic.ObjectListView):
    queryset = IPFabricTransformMap.objects.all()
    table = IPFabricTransformMapTable

    def get_required_permission(self):
        return "ipfabric_netbox.restore_ipfabrictransformmap"

    def get(self, request):
        if request.htmx:
            viewname = get_viewname(self.queryset.model, action="restore")
            form_url = reverse(viewname)
            form = ConfirmationForm(initial=request.GET)
            dependent_objects = {
                IPFabricTransformMap: IPFabricTransformMap.objects.filter(
                    group__isnull=True
                ),
                IPFabricTransformField: IPFabricTransformField.objects.filter(
                    transform_map__group__isnull=True
                ),
                IPFabricRelationshipField: IPFabricRelationshipField.objects.filter(
                    transform_map__group__isnull=True
                ),
            }
            return render(
                request,
                "ipfabric_netbox/ipfabrictransformmap_restore.html",
                {
                    "form": form,
                    "form_url": form_url,
                    "dependent_objects": dependent_objects,
                },
            )
        return redirect(reverse("plugins:ipfabric_netbox:ipfabrictransformmap_list"))

    def post(self, request):
        IPFabricTransformMap.objects.filter(group__isnull=True).delete()
        build_transform_maps(data=get_transform_map())
        return redirect("plugins:ipfabric_netbox:ipfabrictransformmap_list")


@register_model_view(IPFabricTransformMap, "clone")
class IPFabricTransformMapCloneView(BaseObjectView):
    queryset = IPFabricTransformMap.objects.all()
    template_name = "ipfabric_netbox/inc/clone_form.html"
    form = IPFabricTransformMapCloneForm

    def get_required_permission(self):
        return "ipfabric_netbox.clone_ipfabrictransformmap"

    def get(self, request, pk):
        obj = get_object_or_404(self.queryset, pk=pk)
        if request.htmx:
            viewname = get_viewname(self.queryset.model, action="clone")
            form_url = reverse(viewname, kwargs={"pk": obj.pk})
            initial = request.GET.copy()
            initial["name"] = f"Clone of {obj.name}"
            form = self.form(initial=initial)
            restrict_form_fields(form, request.user)
            return render(
                request,
                self.template_name,
                {
                    "object": obj,
                    "form": form,
                    "pk": pk,
                    "form_url": form_url,
                },
            )
        return redirect(obj.get_absolute_url())

    def post(self, request, pk):
        obj = get_object_or_404(self.queryset, pk=pk)
        form = self.form(request.POST)
        restrict_form_fields(form, request.user)
        try:
            if form.is_valid():
                with transaction.atomic():
                    fields = IPFabricTransformField.objects.filter(transform_map=obj)
                    relationships = IPFabricRelationshipField.objects.filter(
                        transform_map=obj
                    )
                    # Clone the transform map - create a proper copy using Django model copying
                    new_map = IPFabricTransformMap(
                        name=form.cleaned_data["name"],
                        source_endpoint=obj.source_endpoint,
                        target_model=obj.target_model,
                        group=form.cleaned_data["group"],
                    )
                    new_map.full_clean()
                    new_map.save()

                    # Clone related transform fields
                    if form.cleaned_data["clone_fields"]:
                        for field in fields:
                            IPFabricTransformField.objects.create(
                                transform_map=new_map,
                                source_field=field.source_field,
                                target_field=field.target_field,
                                coalesce=field.coalesce,
                                template=field.template,
                            )

                    # Clone related relationship fields
                    if form.cleaned_data["clone_relationships"]:
                        for rel in relationships:
                            IPFabricRelationshipField.objects.create(
                                transform_map=new_map,
                                source_model=rel.source_model,
                                target_field=rel.target_field,
                                coalesce=rel.coalesce,
                                template=rel.template,
                            )

                return_url = reverse(
                    "plugins:ipfabric_netbox:ipfabrictransformmap", args=[new_map.pk]
                )
                if request.htmx:
                    response = HttpResponse()
                    response["HX-Redirect"] = return_url
                    return response
                return redirect(return_url)
        except ValidationError as err:
            if not hasattr(err, "error_dict") or not err.error_dict:
                form.add_error(None, err)
            else:
                # This serves to show errors in the form directly
                for field, error in err.error_dict.items():
                    if field in form.fields:
                        form.add_error(field, error)
                    else:
                        form.add_error(None, error)
        if request.htmx:
            viewname = get_viewname(self.queryset.model, action="clone")
            form_url = reverse(viewname, kwargs={"pk": obj.pk})
            response = render(
                request,
                "ipfabric_netbox/inc/clone_form.html",
                {
                    "form": form,
                    "object": obj,
                    "pk": pk,
                    "form_url": form_url,
                },
            )
            response["X-Debug-HTMX-Partial"] = "true"
            return response
        return render(
            request,
            self.template_name,
            {
                "form": form,
                "object": obj,
                "pk": pk,
            },
        )


@register_model_view(IPFabricTransformMap, "fields")
class IPFabricTransformFieldView(generic.ObjectChildrenView):
    queryset = IPFabricTransformMap.objects.all()
    child_model = IPFabricTransformField
    table = IPFabricTransformFieldTable
    template_name = "ipfabric_netbox/inc/transform_map_field_map.html"
    actions = (AddObject, BulkDelete)
    tab = ViewTab(
        label=_("Field Maps"),
        badge=lambda obj: IPFabricTransformField.objects.filter(
            transform_map=obj
        ).count(),
        permission="ipfabric_netbox.view_ipfabrictransformfield",
    )

    def get_children(self, request, parent):
        return self.child_model.objects.filter(transform_map=parent)


@register_model_view(IPFabricTransformMap, "relationships")
class IPFabricTransformRelationshipView(generic.ObjectChildrenView):
    queryset = IPFabricTransformMap.objects.all()
    child_model = IPFabricRelationshipField
    table = IPFabricRelationshipFieldTable
    template_name = "ipfabric_netbox/inc/transform_map_relationship_map.html"
    actions = (AddObject, BulkDelete)
    tab = ViewTab(
        label=_("Relationship Maps"),
        badge=lambda obj: IPFabricRelationshipField.objects.filter(
            transform_map=obj
        ).count(),
        permission="ipfabric_netbox.view_ipfabricrelationshipfield",
    )

    def get_children(self, request, parent):
        return self.child_model.objects.filter(transform_map=parent)


# endregion
# region - Transform Map Field


@register_model_view(IPFabricTransformField, "list", path="", detail=False)
class IPFabricTransformFieldListView(generic.ObjectListView):
    queryset = IPFabricTransformField.objects.all()
    table = IPFabricTransformFieldTable
    actions = ()


@register_model_view(IPFabricTransformField, "add", detail=False)
@register_model_view(IPFabricTransformField, "edit")
class IPFabricTransformFieldEditView(generic.ObjectEditView):
    queryset = IPFabricTransformField.objects.all()
    form = IPFabricTransformFieldForm


@register_model_view(IPFabricTransformField, "delete")
class IPFabricTransformFieldDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricTransformField.objects.all()


@register_model_view(IPFabricTransformField, "bulk_delete", path="delete", detail=False)
class IPFabricTransformFieldBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricTransformField.objects.all()
    table = IPFabricTransformFieldTable


# endregion
# region - Snapshot


@register_model_view(IPFabricSnapshot, "list", path="", detail=False)
class IPFabricSnapshotListView(generic.ObjectListView):
    queryset = IPFabricSnapshot.objects.all()
    table = IPFabricSnapshotTable
    filterset = IPFabricSnapshotFilterSet
    filterset_form = IPFabricSnapshotFilterForm
    actions = (BulkExport, BulkDelete)


@register_model_view(IPFabricSnapshot)
class IPFabricSnapshotView(generic.ObjectView):
    queryset = IPFabricSnapshot.objects.all()


@register_model_view(IPFabricSnapshot, "delete")
class IPFabricSnapshotDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricSnapshot.objects.all()


@register_model_view(IPFabricSnapshot, "bulk_delete", path="delete", detail=False)
class IPFabricSnapshotBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricSnapshot.objects.all()
    filterset = IPFabricSnapshotFilterSet
    table = IPFabricSnapshotTable


@register_model_view(IPFabricSnapshot, "data")
class IPFabricSnapshotRawView(generic.ObjectChildrenView):
    queryset = IPFabricSnapshot.objects.all()
    child_model = IPFabricData
    table = IPFabricDataTable
    template_name = "ipfabric_netbox/inc/snapshotdata.html"
    tab = ViewTab(
        label=_("Raw Data"),
        badge=lambda obj: IPFabricData.objects.filter(snapshot_data=obj).count(),
        permission="ipfabric_netbox.view_ipfabricsnapshot",
        hide_if_empty=True,
    )

    def get_children(self, request, parent):
        return self.child_model.objects.filter(snapshot_data=parent)


# endregion
# region - Snapshot Data


@register_model_view(IPFabricData, "delete")
class IPFabricSnapshotDataDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricData.objects.all()


@register_model_view(IPFabricData, "bulk_delete", path="delete", detail=False)
class IPFabricSnapshotDataBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricData.objects.all()
    filterset = IPFabricDataFilterSet
    table = IPFabricDataTable


@register_model_view(
    IPFabricData,
    name="data",
    path="json",
    kwargs={},
)
class IPFabricSnapshotDataJSONView(generic.ObjectView):
    queryset = IPFabricData.objects.all()
    template_name = "ipfabric_netbox/inc/json.html"

    def get(self, request, **kwargs):
        data = get_object_or_404(IPFabricData, pk=kwargs.get("pk"))
        if request.htmx:
            return render(
                request,
                self.template_name,
                {
                    "object": data,
                },
            )
        return render(
            request,
            self.template_name,
            {
                "object": data,
            },
        )


# endregion
# region - Source


@register_model_view(IPFabricSource, "list", path="", detail=False)
class IPFabricSourceListView(generic.ObjectListView):
    queryset = IPFabricSource.objects.annotate(
        snapshot_count=count_related(IPFabricSnapshot, "source")
    )
    filterset = IPFabricSourceFilterSet
    filterset_form = IPFabricSourceFilterForm
    table = IPFabricSourceTable
    actions = (AddObject, BulkExport, BulkEdit, BulkRename, BulkDelete)


@register_model_view(IPFabricSource, "add", detail=False)
@register_model_view(IPFabricSource, "edit")
class IPFabricSourceEditView(generic.ObjectEditView):
    queryset = IPFabricSource.objects.all()
    form = IPFabricSourceForm


@register_model_view(IPFabricSource)
class IPFabricSourceView(GetRelatedModelsMixin, generic.ObjectView):
    queryset = IPFabricSource.objects.all()

    def get_extra_context(self, request, instance):
        job = instance.jobs.order_by("id").last()
        data = {
            "related_models": self.get_related_models(request, instance),
            "job": job,
        }
        if job:
            data["job_results"] = job.data
        return data


@register_model_view(IPFabricSource, "sync")
class IPFabricSourceSyncView(BaseObjectView):
    queryset = IPFabricSource.objects.all()

    def get_required_permission(self):
        return "ipfabric_netbox.sync_ipfabricsource"

    def get(self, request, pk):
        ipfabricsource = get_object_or_404(self.queryset, pk=pk)
        return redirect(ipfabricsource.get_absolute_url())

    def post(self, request, pk):
        ipfabricsource = get_object_or_404(self.queryset, pk=pk)
        job = ipfabricsource.enqueue_sync_job(request=request)

        messages.success(request, f"Queued job #{job.pk} to sync {ipfabricsource}")
        return redirect(ipfabricsource.get_absolute_url())


@register_model_view(IPFabricSource, "delete")
class IPFabricSourceDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricSource.objects.all()


@register_model_view(IPFabricSource, "bulk_edit", path="edit", detail=False)
class IPFabricSourceBulkEditView(generic.BulkEditView):
    queryset = IPFabricSource.objects.all()
    table = IPFabricSourceTable
    form = IPFabricSourceBulkEditForm


@register_model_view(IPFabricSource, "bulk_rename", path="rename", detail=False)
class IPFabricSourceBulkRenameView(generic.BulkRenameView):
    queryset = IPFabricSource.objects.all()


@register_model_view(IPFabricSource, "bulk_delete", path="delete", detail=False)
class IPFabricSourceBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricSource.objects.all()
    filterset = IPFabricSourceFilterSet
    table = IPFabricSourceTable


@register_model_view(
    IPFabricSource,
    name="topology",
    path="topology/<int:site>",
    kwargs={"snapshot": ""},
)
class IPFabricSourceTopology(LoginRequiredMixin, View):
    template_name = "ipfabric_netbox/inc/site_topology_modal.html"

    def get(self, request, pk, site, **kwargs):
        if request.htmx:
            try:
                site = get_object_or_404(Site, pk=site)
                source_id = request.GET.get("source")
                if not source_id:
                    raise Exception("Source ID not available in request.")
                source = get_object_or_404(IPFabricSource, pk=source_id)
                snapshot = request.GET.get("snapshot")
                if not snapshot:
                    raise Exception("Snapshot ID not available in request.")

                source.parameters.update(
                    {"snapshot_id": snapshot, "base_url": source.url}
                )

                ipf = IPFabric(parameters=source.parameters)
                snapshot_data = ipf.ipf.snapshots.get(snapshot)
                if not snapshot_data:
                    raise Exception(
                        f"Snapshot ({snapshot}) not available in IP Fabric."  # noqa E713
                    )

                sites = ipf.ipf.inventory.sites.all(
                    filters={"siteName": ["eq", site.name]}
                )
                if not sites:
                    raise Exception(
                        f"{site.name} not available in snapshot ({snapshot})."  # noqa E713
                    )

                net = Network(sites=site.name, all_network=False)
                settings = NetworkSettings()
                settings.hide_protocol("xdp")
                settings.hiddenDeviceTypes.extend(["transit", "cloud"])

                link = ipf.ipf.diagram.share_link(net, graph_settings=settings)
                svg_data = ipf.ipf.diagram.svg(net, graph_settings=settings).decode(
                    "utf-8"
                )
                error = None
            except Exception as e:
                error = e
                svg_data = link = snapshot_data = source = None

            return render(
                request,
                self.template_name,
                {
                    "site": site,
                    "source": source,
                    "svg": svg_data,
                    "size": "xl",
                    "link": link,
                    "time": timezone.now(),
                    "snapshot": snapshot_data,
                    "error": error,
                },
            )
        return render(
            request,
            self.template_name,
            {
                "site": site,
                "size": "xl",
                "time": timezone.now(),
            },
        )


# endregion
# region - Sync
@register_model_view(IPFabricSync, "list", path="", detail=False)
class IPFabricSyncListView(generic.ObjectListView):
    queryset = IPFabricSync.objects.annotate(
        last_ingestion_pk=models.Subquery(
            IPFabricIngestion.objects.filter(sync=models.OuterRef("pk"))
            .order_by("-pk")
            .values("pk")[:1]
        )
    )
    table = IPFabricSyncTable
    filterset = IPFabricSyncFilterSet
    actions = (AddObject, BulkExport, BulkEdit, BulkRename, BulkDelete)


@register_model_view(IPFabricSync, "add", detail=False)
@register_model_view(IPFabricSync, "edit")
class IPFabricSyncEditView(generic.ObjectEditView):
    queryset = IPFabricSync.objects.all()
    form = IPFabricSyncForm

    def alter_object(self, obj, request, url_args, url_kwargs):
        obj.user = request.user
        return obj


@register_model_view(IPFabricSync)
class IPFabricSyncView(generic.ObjectView):
    queryset = IPFabricSync.objects.all()

    def get(self, request, **kwargs):
        # Handle HTMX requests separately
        if request.htmx:
            instance = self.get_object(**kwargs)
            last_ingestion = instance.ipfabricingestion_set.last()

            response = render(
                request,
                "ipfabric_netbox/partials/sync_last_ingestion.html",
                {"last_ingestion": last_ingestion},
            )

            if instance.status not in ["queued", "syncing"]:
                messages.success(
                    request,
                    f"Ingestion ({instance.name}) {instance.status}. Ingestion {last_ingestion.name} {last_ingestion.job.status}.",
                )
                response["HX-Refresh"] = "true"
            return response

        # For regular requests, use the parent method which includes actions
        return super().get(request, **kwargs)

    def get_extra_context(self, request, instance):
        if request.GET.get("format") in ["json", "yaml"]:
            format = request.GET.get("format")
            if request.user.is_authenticated:
                request.user.config.set("data_format", format, commit=True)
        elif request.user.is_authenticated:
            format = request.user.config.get("data_format", "json")
        else:
            format = "json"

        last_ingestion = instance.ipfabricingestion_set.last()

        scheduled_job = None
        if instance.scheduled:
            scheduled_job = instance.jobs.filter(scheduled=instance.scheduled).last()

        return {
            "format": format,
            "last_ingestion": last_ingestion,
            "scheduled_job": scheduled_job,
        }


@register_model_view(IPFabricSync, "sync")
class IPFabricStartSyncView(BaseObjectView):
    queryset = IPFabricSync.objects.all()

    def get_required_permission(self):
        return "ipfabric_netbox.sync_ipfabricsync"

    def get(self, request, pk):
        ipfabric = get_object_or_404(self.queryset, pk=pk)
        return redirect(ipfabric.get_absolute_url())

    def post(self, request, pk):
        ipfabric = get_object_or_404(self.queryset, pk=pk)
        job = ipfabric.enqueue_sync_job(user=request.user, adhoc=True)

        messages.success(request, f"Queued job #{job.pk} to sync {ipfabric}")
        return redirect(ipfabric.get_absolute_url())


@register_model_view(IPFabricSync, "delete")
class IPFabricSyncDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricSync.objects.all()
    default_return_url = "plugins:ipfabric_netbox:ipfabricsync_list"


@register_model_view(IPFabricSync, "bulk_edit", path="edit", detail=False)
class IPFabricSyncBulkEditView(generic.BulkEditView):
    queryset = IPFabricSync.objects.all()
    table = IPFabricSyncTable
    form = IPFabricSyncBulkEditForm


@register_model_view(IPFabricSync, "bulk_rename", path="rename", detail=False)
class IPFabricSyncBulkRenameView(generic.BulkRenameView):
    queryset = IPFabricSync.objects.all()


@register_model_view(IPFabricSync, "bulk_delete", path="delete", detail=False)
class IPFabricSyncBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricSync.objects.all()
    filterset = IPFabricSnapshotFilterSet
    table = IPFabricSyncTable


@register_model_view(IPFabricSync, "transformmaps")
class IPFabricTransformMapSyncTabView(generic.ObjectChildrenView):
    queryset = IPFabricSync.objects.all()
    child_model = IPFabricTransformMap
    table = IPFabricTransformMapTable
    template_name = "generic/object_children.html"
    actions = (AddObject, BulkDelete)
    tab = ViewTab(
        label=_("Transform Maps"),
        badge=lambda obj: obj.get_transform_maps(
            obj.parameters.get("groups", []) if obj.parameters else []
        ).count(),
        permission="ipfabric_netbox.view_ipfabrictransformmap",
    )

    def get_children(self, request, parent):
        return parent.get_transform_maps(
            parent.parameters.get("groups", []) if parent.parameters else []
        )


@register_model_view(IPFabricSync, "filters")
class IPFabricTransformMapFilterTabView(generic.ObjectChildrenView):
    queryset = IPFabricSync.objects.all()
    child_model = IPFabricFilter
    table = IPFabricFilterTable
    template_name = "generic/object_children.html"
    actions = (AddObject, BulkDelete)
    tab = ViewTab(
        label=_("Filters"),
        badge=lambda obj: obj.filters.count(),
        permission="ipfabric_netbox.view_ipfabricfilter",
    )

    def get_children(self, request, parent):
        return parent.filters.all()


@register_model_view(IPFabricSync, "endpoints")
class IPFabricSyncEndpointTabView(generic.ObjectChildrenView):
    queryset = IPFabricSync.objects.all()
    child_model = IPFabricEndpoint
    table = IPFabricEndpointTable
    template_name = "generic/object_children.html"
    actions = ()
    tab = ViewTab(
        label=_("Endpoints"),
        badge=lambda obj: IPFabricEndpoint.objects.filter(filters__syncs=obj)
        .distinct()
        .count(),
        permission="ipfabric_netbox.view_ipfabricendpoint",
    )

    def get_children(self, request, parent):
        # Get all unique endpoints used by filters in this sync
        return (
            IPFabricEndpoint.objects.filter(filters__syncs=parent)
            .distinct()
            .annotate(
                sync_pk=models.Value(parent.pk, output_field=models.IntegerField()),
                filters_count=models.Count(
                    "filters", filter=models.Q(filters__syncs=parent)
                ),
            )
        )

    def get_table(self, *args, **kwargs):
        table = super().get_table(*args, **kwargs)
        # Override default columns to show the context-specific ones
        table.default_columns = ("name", "endpoint", "filters_count", "show_filters")
        return table


@register_model_view(
    IPFabricSync,
    name="endpoint_filters",
    path="endpoint/<int:endpoint_pk>/filters",
    kwargs={"model": IPFabricSync},
)
class IPFabricSyncEndpointFiltersView(CombinedExpressionsBaseView):
    """Display combined filters for an endpoint within a specific sync."""

    def get_context_data(self, request, **kwargs):
        """Provide context for endpoint filters within a sync.

        Args:
            request: HTTP request (must be HTMX)
            **kwargs: Must contain 'pk' (sync primary key) and 'endpoint_pk'

        Returns:
            dict: Standardized context with endpoint, sync, and combined filters
        """
        sync_pk = kwargs.get("pk")
        endpoint_pk = kwargs.get("endpoint_pk")

        # Return empty context if not HTMX or missing required parameters
        if not request.htmx or not sync_pk or not endpoint_pk:
            return {
                "object": None,
                "combined_expressions": {},
                "is_empty": True,
                "context_type": "endpoint",
            }

        # Fetch sync and endpoint objects
        sync_obj = IPFabricSync.objects.get(pk=sync_pk)
        endpoint_obj = IPFabricEndpoint.objects.get(pk=endpoint_pk)

        # Combine filters for this endpoint within this sync
        combined = endpoint_obj.combine_filters(sync=sync_obj)

        return {
            "object": endpoint_obj,  # The endpoint is the main object
            "sync": sync_obj,  # Pass sync for template context
            "combined_expressions": combined,
            "is_empty": len(combined) == 0,
            "context_type": "endpoint",
        }


@register_model_view(IPFabricSync, "ingestion")
class IPFabricIngestionTabView(generic.ObjectChildrenView):
    queryset = IPFabricSync.objects.all()
    child_model = IPFabricIngestion
    table = IPFabricIngestionTable
    filterset = IPFabricIngestionFilterSet
    actions = (AddObject, BulkDelete)
    tab = ViewTab(
        label=_("Ingestions"),
        badge=lambda obj: IPFabricIngestion.objects.filter(sync=obj).count(),
        permission="ipfabric_netbox.view_ipfabricingestion",
    )

    def get_children(self, request, parent):
        return self.child_model.objects.filter(sync=parent).annotate(
            description=models.F("branch__description"),
            user=models.F("sync__user__username"),
            staged_changes=models.Count(models.F("branch__changediff")),
            branch_name=models.F("branch__name"),
            sync_name=models.F("sync__name"),
        )


# endregion
# region - Ingestion


@register_model_view(IPFabricIngestion, "list", path="", detail=False)
class IPFabricIngestionListView(generic.ObjectListView):
    queryset = IPFabricIngestion.objects.annotate(
        description=models.F("branch__description"),
        user=models.F("sync__user__username"),
        staged_changes=models.Count(models.F("branch__changediff")),
        branch_name=models.F("branch__name"),
        sync_name=models.F("sync__name"),
    )
    filterset = IPFabricIngestionFilterSet
    filterset_form = IPFabricIngestionFilterForm
    table = IPFabricIngestionTable
    actions = (AddObject, BulkExport, BulkDelete)


def annotate_statistics(queryset):
    return queryset.annotate(
        num_created=models.Count(
            "branch__changediff",
            filter=models.Q(
                branch__changediff__action=ObjectChangeActionChoices.ACTION_CREATE
            )
            & ~models.Q(branch__changediff__object_type__model="objectchange"),
        ),
        num_updated=models.Count(
            "branch__changediff",
            filter=models.Q(
                branch__changediff__action=ObjectChangeActionChoices.ACTION_UPDATE
            )
            & ~models.Q(branch__changediff__object_type__model="objectchange"),
        ),
        num_deleted=models.Count(
            "branch__changediff",
            filter=models.Q(
                branch__changediff__action=ObjectChangeActionChoices.ACTION_DELETE
            )
            & ~models.Q(branch__changediff__object_type__model="objectchange"),
        ),
        description=models.F("branch__description"),
        user=models.F("sync__user__username"),
        staged_changes=models.Count(models.F("branch__changediff")),
    )


@register_model_view(
    IPFabricIngestion,
    name="logs",
    path="logs",
)
class IPFabricIngestionLogView(LoginRequiredMixin, View):
    template_name = "ipfabric_netbox/partials/ingestion_all.html"

    def get(self, request, **kwargs):
        ingestion_id = kwargs.get("pk")
        ingestion = annotate_statistics(IPFabricIngestion.objects).get(pk=ingestion_id)
        data = ingestion.get_statistics()
        data["object"] = ingestion
        data["job"] = ingestion.job

        if request.htmx:
            response = render(
                request,
                self.template_name,
                data,
            )
            if not ingestion.job or ingestion.job.completed:
                response["HX-Refresh"] = "true"
            return response
        return render(request, self.template_name, data)


@register_model_view(IPFabricIngestion)
class IPFabricIngestionView(generic.ObjectView):
    queryset = annotate_statistics(IPFabricIngestion.objects)

    def get_extra_context(self, request, instance):
        data = instance.get_statistics()
        return data


@register_model_view(IPFabricIngestion, "merge")
class IPFabricIngestionMergeView(BaseObjectView):
    queryset = IPFabricIngestion.objects.annotate(
        description=models.F("branch__description"),
        user=models.F("sync__user__username"),
        staged_changes=models.Count(models.F("branch__changediff")),
    )
    template_name = "ipfabric_netbox/inc/merge_form.html"
    form = IPFabricIngestionMergeForm

    def get_required_permission(self):
        return "ipfabric_netbox.merge_ipfabricingestion"

    def get(self, request, pk):
        obj = get_object_or_404(self.queryset, pk=pk)

        if request.htmx:
            viewname = get_viewname(self.queryset.model, action="merge")
            form_url = reverse(viewname, kwargs={"pk": obj.pk})
            form = self.form(initial=request.GET)
            restrict_form_fields(form, request.user)
            return render(
                request,
                "ipfabric_netbox/inc/merge_form.html",
                {
                    "object": obj,
                    "object_type": self.queryset.model._meta.verbose_name,
                    "form": form,
                    "form_url": form_url,
                    **self.get_extra_context(request, obj),
                },
            )

        return redirect(obj.get_absolute_url())

    def post(self, request, pk):
        ingestion = get_object_or_404(self.queryset, pk=pk)
        form = self.form(request.POST)
        restrict_form_fields(form, request.user)
        if form.is_valid():
            job = ingestion.enqueue_merge_job(
                user=request.user, remove_branch=form.cleaned_data["remove_branch"]
            )
            messages.success(request, f"Queued job #{job.pk} to sync {ingestion}")
            return redirect(ingestion.get_absolute_url())

        # Handle invalid form - add form errors to messages and redirect back
        for field, errors in form.errors.items():
            for error in errors:
                messages.error(request, f"{field}: {error}")
        if form.non_field_errors():
            for error in form.non_field_errors():
                messages.error(request, error)

        return redirect(ingestion.get_absolute_url())


@register_model_view(
    IPFabricIngestion,
    name="change_diff",
    path="change/<int:change_pk>",
    kwargs={"model": IPFabricIngestion},
)
class IPFabricIngestionChangesDiffView(LoginRequiredMixin, View):
    template_name = "ipfabric_netbox/inc/diff.html"

    def get(self, request, **kwargs):
        change_id = kwargs.get("change_pk", None)

        if not request.htmx or not change_id:
            return render(
                request,
                self.template_name,
                {
                    "change": None,
                    "prechange_data": None,
                    "postchange_data": None,
                    "diff_added": None,
                    "diff_removed": None,
                    "size": "lg",
                },
            )

        change = ChangeDiff.objects.get(pk=change_id)
        if change.original and change.modified:
            diff_added = shallow_compare_dict(
                change.original or dict(),
                change.modified or dict(),
                exclude=["last_updated"],
            )
            diff_removed = (
                {x: change.original.get(x) for x in diff_added}
                if change.modified
                else {}
            )
        else:
            diff_added = None
            diff_removed = None

        return render(
            request,
            self.template_name,
            {
                "change": change,
                "prechange_data": change.original,
                "postchange_data": change.modified,
                "diff_added": diff_added,
                "diff_removed": diff_removed,
                "size": "lg",
            },
        )


@register_model_view(IPFabricIngestion, "change")
class IPFabricIngestionChangesView(generic.ObjectChildrenView):
    queryset = IPFabricIngestion.objects.all()
    child_model = ChangeDiff
    table = IPFabricIngestionChangesTable
    filterset = IPFabricIngestionChangeFilterSet
    template_name = "generic/object_children.html"
    tab = ViewTab(
        label=_("Changes"),
        badge=lambda obj: ChangeDiff.objects.filter(branch=obj.branch).count(),
        permission="ipfabric_netbox.view_ipfabricingestion",
    )

    def get_children(self, request, parent):
        return self.child_model.objects.filter(branch=parent.branch)


@register_model_view(IPFabricIngestion, "ingestion_issues")
class IPFabricIngestionIssuesView(generic.ObjectChildrenView):
    queryset = IPFabricIngestion.objects.all()
    child_model = IPFabricIngestionIssue
    table = IPFabricIngestionIssuesTable
    template_name = "generic/object_children.html"
    filterset = IPFabricIngestionIssueFilterSet
    actions = (AddObject, BulkDelete)
    tab = ViewTab(
        label=_("Ingestion Issues"),
        badge=lambda obj: IPFabricIngestionIssue.objects.filter(ingestion=obj).count(),
        permission="ipfabric_netbox.view_ipfabricingestionissue",
    )

    def get_children(self, request, parent):
        return IPFabricIngestionIssue.objects.filter(ingestion=parent)


@register_model_view(IPFabricIngestion, "delete")
class IPFabricIngestionDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricIngestion.objects.all()


@register_model_view(IPFabricIngestion, "bulk_delete", path="delete", detail=False)
class IPFabricIngestionBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricIngestion.objects.all()
    table = IPFabricIngestionTable


@register_model_view(Device, "ipfabric")
class IPFabricTable(generic.ObjectView):
    template_name = "ipfabric_netbox/ipfabric_table.html"
    tab = ViewTab("IP Fabric", permission="ipfabric_netbox.view_devicetable")
    queryset = Device.objects.all()

    def get_extra_context(self, request, instance):
        """Process form and prepare table data for the template."""
        device = instance
        form = (
            IPFabricTableForm(request.GET)
            if "table" in request.GET
            else IPFabricTableForm()
        )
        restrict_form_fields(form, request.user)
        data = None
        source = None

        if form.is_valid():
            table_name = form.cleaned_data["table"]
            test = {
                "True": True,
                "False": False,
            }
            cache_enable = test.get(form.cleaned_data["cache_enable"])
            source = form.cleaned_data.get("source")

            if not form.cleaned_data["snapshot_data"]:
                snapshot_id = "$last"
                source = (
                    source
                    or IPFabricSource.objects.filter(
                        pk=device.custom_field_data.get("ipfabric_source")
                    ).first()
                    or IPFabricSource.get_for_site(device.site).first()
                )
            else:
                snapshot_id = form.cleaned_data["snapshot_data"].snapshot_id
                source = source or form.cleaned_data["snapshot_data"].source

            if source is not None:
                source.parameters["snapshot_id"] = snapshot_id
                source.parameters["base_url"] = source.url

                cache_key = f"ipfabric_{table_name}_{device.serial}_{source.parameters['snapshot_id']}"
                if cache_enable:
                    data = cache.get(cache_key)

                if not data:
                    try:
                        ipf = IPFabric(parameters=source.parameters)
                        raw_data, columns = ipf.get_table_data(
                            table=table_name, device=device
                        )
                        data = {"data": raw_data, "columns": columns}
                        cache.set(cache_key, data, 60 * 60 * 24)
                    except Exception as e:
                        messages.error(request, e)

        if not data:
            data = {"data": [], "columns": []}

        table = DeviceIPFTable(data["data"], extra_columns=data["columns"])

        RequestConfig(
            request,
            {
                "paginator_class": EnhancedPaginator,
                "per_page": get_paginate_count(request),
            },
        ).configure(table)

        if not source:
            if source_id := device.custom_field_data.get("ipfabric_source"):
                source = IPFabricSource.objects.filter(pk=source_id).first()
            else:
                source = IPFabricSource.get_for_site(device.site).first()

        return {
            "source": source,
            "form": form,
            "table": table,
        }

    def get(self, request, **kwargs):
        """Handle GET requests, with special handling for HTMX table updates."""
        # For HTMX requests, we only need to return the table HTML
        if request.htmx:
            device = get_object_or_404(Device, pk=kwargs.get("pk"))
            context = self.get_extra_context(request, device)
            return render(
                request,
                "htmx/table.html",
                {
                    "table": context["table"],
                },
            )

        # For regular requests, use the parent's get() method which will call get_extra_context()
        return super().get(request, **kwargs)


# endregion


# region - Filter
@register_model_view(IPFabricFilter, "list", path="", detail=False)
class IPFabricFilterListView(generic.ObjectListView):
    queryset = IPFabricFilter.objects.all()
    table = IPFabricFilterTable
    filterset = IPFabricFilterFilterSet
    template_name = "ipfabric_netbox/htmx_list.html"


@register_model_view(IPFabricFilter, "add", detail=False)
@register_model_view(IPFabricFilter, "edit")
class IPFabricFilterEditView(generic.ObjectEditView):
    queryset = IPFabricFilter.objects.all()
    form = IPFabricFilterForm
    default_return_url = "plugins:ipfabric_netbox:ipfabricfilter_list"


@register_model_view(IPFabricFilter, "delete")
class IPFabricFilterDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricFilter.objects.all()
    default_return_url = "plugins:ipfabric_netbox:ipfabricfilter_list"


@register_model_view(IPFabricFilter, "bulk_import", path="import", detail=False)
class IPFabricFilterBulkImportView(generic.BulkImportView):
    queryset = IPFabricFilter.objects.all()
    model_form = IPFabricFilterBulkImportForm


@register_model_view(IPFabricFilter, "bulk_edit", path="edit", detail=False)
class IPFabricFilterBulkEditView(generic.BulkEditView):
    queryset = IPFabricFilter.objects.all()
    table = IPFabricFilterTable
    form = IPFabricFilterBulkEditForm


@register_model_view(IPFabricFilter, "bulk_rename", path="rename", detail=False)
class IPFabricFilterBulkRenameView(generic.BulkRenameView):
    queryset = IPFabricFilter.objects.all()


@register_model_view(IPFabricFilter, "bulk_delete", path="delete", detail=False)
class IPFabricFilterBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricFilter.objects.all()
    table = IPFabricFilterTable


@register_model_view(IPFabricFilter)
class IPFabricFilterView(GetRelatedModelsMixin, generic.ObjectView):
    queryset = IPFabricFilter.objects.all()

    def get_extra_context(self, request, instance):
        return {
            "related_models": self.get_related_models(
                request,
                instance,
                extra=(
                    (
                        IPFabricSync.objects.restrict(request.user, "view").filter(
                            filters__in=[instance]
                        ),
                        "filter_id",
                    ),
                    (
                        IPFabricEndpoint.objects.restrict(request.user, "view").filter(
                            filters__in=[instance]
                        ),
                        "ipfabric_filter_id",
                    ),
                    (
                        IPFabricFilterExpression.objects.restrict(
                            request.user, "view"
                        ).filter(filters__in=[instance]),
                        "ipfabric_filter_id",
                    ),
                ),
            ),
        }


@register_model_view(
    IPFabricFilter,
    name="combined_expressions",
    path="combined-expressions",
    kwargs={"model": IPFabricFilter},
)
class IPFabricFilterCombinedExpressionsView(CombinedExpressionsBaseView):
    """Display merged expressions for a single filter."""

    def get_context_data(self, request, **kwargs):
        """Provide context for filter's merged expressions.

        Args:
            request: HTTP request (must be HTMX)
            **kwargs: Must contain 'pk' (filter primary key)

        Returns:
            dict: Standardized context with filter object and merged expressions
        """
        filter_pk = kwargs.get("pk")

        # Return empty context if not HTMX or missing required parameters
        if not request.htmx or not filter_pk:
            return {
                "object": None,
                "combined_expressions": [],
                "is_empty": True,
                "context_type": "filter",
            }

        # Force fresh query from database, prefetch expressions
        filter_obj = IPFabricFilter.objects.prefetch_related("expressions").get(
            pk=filter_pk
        )
        combined = filter_obj.merge_expressions()

        return {
            "object": filter_obj,
            "combined_expressions": combined,
            "is_empty": len(combined) == 0,
            "context_type": "filter",
        }


# endregion


# region - Filter Expression
@register_model_view(IPFabricFilterExpression, "list", path="", detail=False)
class IPFabricFilterExpressionListView(generic.ObjectListView):
    queryset = IPFabricFilterExpression.objects.all()
    table = IPFabricFilterExpressionTable
    filterset = IPFabricFilterExpressionFilterSet


@register_model_view(IPFabricFilterExpression, "add", detail=False)
@register_model_view(IPFabricFilterExpression, "edit")
class IPFabricFilterExpressionEditView(generic.ObjectEditView):
    queryset = IPFabricFilterExpression.objects.all()
    form = IPFabricFilterExpressionForm
    default_return_url = "plugins:ipfabric_netbox:ipfabricfilterexpression_list"
    template_name = "ipfabric_netbox/ipfabricfilterexpression_edit.html"

    def get_extra_context(self, request, instance):
        """Add test configuration context to the template."""
        context = super().get_extra_context(request, instance)

        # Add test-related context if editing existing object
        if instance.pk:
            # Get available test sources and endpoints
            sources, endpoints = get_filter_expression_test_candidates(instance)
            context["available_test_sources"] = list(sources)
            context["available_test_endpoints"] = list(endpoints)
            context["can_test"] = bool(sources and endpoints)
        else:
            context["can_test"] = False

        return context


@register_model_view(IPFabricFilterExpression, "delete")
class IPFabricFilterExpressionDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricFilterExpression.objects.all()
    default_return_url = "plugins:ipfabric_netbox:ipfabricfilterexpression_list"


@register_model_view(
    IPFabricFilterExpression, "bulk_import", path="import", detail=False
)
class IPFabricFilterExpressionBulkImportView(generic.BulkImportView):
    queryset = IPFabricFilterExpression.objects.all()
    model_form = IPFabricFilterExpressionBulkImportForm


@register_model_view(IPFabricFilterExpression, "bulk_edit", path="edit", detail=False)
class IPFabricFilterExpressionBulkEditView(generic.BulkEditView):
    queryset = IPFabricFilterExpression.objects.all()
    table = IPFabricFilterExpressionTable
    form = IPFabricFilterExpressionBulkEditForm


@register_model_view(
    IPFabricFilterExpression, "bulk_rename", path="rename", detail=False
)
class IPFabricFilterExpressionBulkRenameView(generic.BulkRenameView):
    queryset = IPFabricFilterExpression.objects.all()


@register_model_view(
    IPFabricFilterExpression, "bulk_delete", path="delete", detail=False
)
class IPFabricFilterExpressionBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricFilterExpression.objects.all()
    table = IPFabricFilterExpressionTable


@register_model_view(IPFabricFilterExpression)
class IPFabricFilterExpressionView(GetRelatedModelsMixin, generic.ObjectView):
    queryset = IPFabricFilterExpression.objects.all()

    def get_extra_context(self, request, instance):
        return {
            "related_models": self.get_related_models(
                request,
                instance,
                extra=(
                    (
                        IPFabricFilter.objects.restrict(request.user, "view").filter(
                            expressions__in=[instance]
                        ),
                        "expression_id",
                    ),
                ),
            ),
        }


@register_model_view(IPFabricFilterExpression, "test")
class IPFabricFilterExpressionTestView(View):
    """Test a filter expression against IP Fabric API."""

    def post(self, request, pk=None):
        # Get test parameters from POST data
        test_source_id = request.POST.get("test_source")
        test_endpoint_id = request.POST.get("test_endpoint")
        expression_json = request.POST.get("expression")

        return self._test_expression_logic(
            request, test_source_id, test_endpoint_id, expression_json
        )

    @staticmethod
    def _test_expression_logic(
        request, test_source_id, test_endpoint_id, expression_json
    ):
        """Shared logic for testing filter expressions (works for both saved and unsaved)."""

        # Validate required fields
        if not test_source_id:
            return HttpResponse(
                json.dumps({"success": False, "error": "Please select a Test Source."}),
                content_type="application/json",
            )

        if not test_endpoint_id:
            return HttpResponse(
                json.dumps(
                    {"success": False, "error": "Please select a Test Endpoint."}
                ),
                content_type="application/json",
            )

        # Validate expression JSON
        if expression_json:
            try:
                expression_data = json.loads(expression_json)
                if not isinstance(expression_data, list):
                    return HttpResponse(
                        json.dumps(
                            {
                                "success": False,
                                "error": "Expression must be a JSON list.",
                            }
                        ),
                        content_type="application/json",
                    )
                for idx, item in enumerate(expression_data):
                    if not isinstance(item, dict):
                        return HttpResponse(
                            json.dumps(
                                {
                                    "success": False,
                                    "error": f"Expression item at index {idx} must be a dictionary.",
                                }
                            ),
                            content_type="application/json",
                        )
            except json.JSONDecodeError as e:
                return HttpResponse(
                    json.dumps({"success": False, "error": f"Invalid JSON: {str(e)}"}),
                    content_type="application/json",
                )
        else:
            return HttpResponse(
                json.dumps({"success": False, "error": "Expression is required."}),
                content_type="application/json",
            )

        # Get source and endpoint objects
        try:
            source = IPFabricSource.objects.get(pk=test_source_id)
        except IPFabricSource.DoesNotExist:
            return HttpResponse(
                json.dumps({"success": False, "error": "Selected source not found."}),
                content_type="application/json",
                status=404,
            )

        # Check if source is LOCAL
        if source.type != IPFabricSourceTypeChoices.LOCAL:
            return HttpResponse(
                json.dumps(
                    {
                        "success": False,
                        "error": "Cannot test against REMOTE sources. Please select a LOCAL IP Fabric source with direct API access.",
                    }
                ),
                content_type="application/json",
            )

        try:
            endpoint = IPFabricEndpoint.objects.get(pk=test_endpoint_id)
        except IPFabricEndpoint.DoesNotExist:
            return HttpResponse(
                json.dumps({"success": False, "error": "Selected endpoint not found."}),
                content_type="application/json",
                status=404,
            )

        # Test the expression against IP Fabric API
        try:
            # Check that source has auth token
            auth = source.parameters.get("auth")
            if not auth:
                return HttpResponse(
                    json.dumps(
                        {
                            "success": False,
                            "error": "Source is missing API authentication token.",
                        }
                    ),
                    content_type="application/json",
                )

            # Prepare parameters for IPFabric client (same pattern as other views)
            parameters = source.parameters.copy()
            parameters.update(
                {
                    "snapshot_id": "$last",
                    "base_url": source.url,
                    "pagination": {"limit": 1, "start": 0},
                }
            )
            # Build filter - use AND type as default (IPF doesn't distinguish between types at this level)
            filter_dict = {"and": expression_data}
            ipf = IPFabric(parameters=parameters)
            results = ipf.ipf.fetch_all(
                url=endpoint.endpoint, snapshot_id="$last", filters=filter_dict
            )

            # Return success
            return HttpResponse(
                json.dumps(
                    {
                        "success": True,
                        "count": len(results),
                        "message": f"Test successful! Expression matched {len(results)} result(s) (limited to 1 for testing).",
                    }
                ),
                content_type="application/json",
            )

        except Exception as e:
            # Return error with helpful context
            error_msg = str(e)

            # Enhance error message for common issues
            if "Unrecognized key(s)" in error_msg:
                error_msg += f" | Hint: The field name in your expression doesn't exist in the '{endpoint.endpoint}' endpoint. Check the IP Fabric API documentation for valid field names for this table."
            elif "Invalid Input" in error_msg or "VALIDATION_FAILED" in error_msg:
                error_msg += f" | Hint: Check that your filter syntax is correct and field names are valid for endpoint '{endpoint.endpoint}'."

            return HttpResponse(
                json.dumps({"success": False, "error": f"API Error: {error_msg}"}),
                content_type="application/json",
            )


# endregion



# Generated by Django 5.2 on 2025-05-26 08:53
import django.db.models.deletion
import taggit.managers
import utilities.json
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("extras", "0128_tableconfig"),
        ("ipfabric_netbox", "0001_initial_squashed_0013_switch_to_branching_plugin"),
    ]

    operations = [
        migrations.CreateModel(
            name="IPFabricTransformMapGroup",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=utilities.json.CustomFieldJSONEncoder,
                    ),
                ),
                ("name", models.CharField(max_length=100, unique=True)),
                ("description", models.TextField(blank=True, null=True)),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
            ],
            options={
                "verbose_name": "IP Fabric Transform Map Group",
                "verbose_name_plural": "IP Fabric Transform Map Groups",
            },
        ),
        migrations.AddField(
            model_name="ipfabrictransformmap",
            name="group",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="transform_maps",
                to="ipfabric_netbox.ipfabrictransformmapgroup",
            ),
        ),
        migrations.AlterField(
            model_name="ipfabrictransformmap",
            name="name",
            field=models.CharField(max_length=100),
        ),
    ]



import logging
from collections import OrderedDict
from typing import TYPE_CHECKING

from django import template
from django.apps import apps

if TYPE_CHECKING:
    from ipfabric_netbox.models import IPFabricSync

register = template.Library()

logger = logging.getLogger("ipfabric_netbox.helpers")


@register.filter()
def resolve_object(pk, model_path):
    """
    Usage: {{ pk|resolve_object:"app_label.ModelName" }}
    """
    try:
        app_label, model_name = model_path.split(".")
        model = apps.get_model(app_label, model_name)
        return model.objects.get(pk=pk)
    except Exception:
        return None


@register.filter()
def sort_parameters_hierarchical(
    parameters_dict: dict, sync_obj: "IPFabricSync | None" = None
) -> list:
    """
    Sort parameters hierarchically for IPFabricSync objects:
    1. 'groups' first
    2. Models in hierarchical order from IPFabricSync.get_model_hierarchy()

    For other objects, returns items unsorted.

    Usage: {{ object.parameters|sort_parameters_hierarchical:object }}
    """
    if not parameters_dict or not isinstance(parameters_dict, dict):
        return []

    try:
        group_ids = parameters_dict.get("groups", [])

        # Get the hierarchical model order from IPFabricSync (returns transform maps)
        transform_maps = sync_obj.__class__.get_model_hierarchy(group_ids)
        # Convert transform maps to app_label.model format
        hierarchy_order = [
            f"{tm.target_model.app_label}.{tm.target_model.model}"
            for tm in transform_maps
        ]

        # Group models by app label while maintaining hierarchical order
        app_models = OrderedDict()
        for model_name in hierarchy_order:
            if "." in model_name:
                app_label = model_name.split(".")[0]
                if app_label not in app_models:
                    app_models[app_label] = []
                app_models[app_label].append(model_name)

        # Build sorted list: groups first, then all app models
        sorted_keys = []
        for app_label, models in app_models.items():
            sorted_keys.extend(models)

        result = []
        # Start with keys that weren't in the hierarchy
        for key, value in parameters_dict.items():
            if key not in sorted_keys:
                result.append((key, value))

        # Return items in the sorted order
        for key in sorted_keys:
            if key in parameters_dict:
                result.append((key, parameters_dict[key]))

        return result
    except Exception:
        # Fallback to simple alphabetical sort if something goes wrong
        logger.warning("Failed to sort parameters hierarchically", exc_info=True)
        return sorted(parameters_dict.items())



# Generated by Django 5.2.5 on 2025-08-21 11:21
import taggit.managers
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("extras", "0128_tableconfig"),
        ("ipfabric_netbox", "0015_ipfabricingestionissue"),
    ]

    operations = [
        migrations.AddField(
            model_name="ipfabricsnapshot",
            name="tags",
            field=taggit.managers.TaggableManager(
                through="extras.TaggedItem", to="extras.Tag"
            ),
        ),
        migrations.AlterField(
            model_name="ipfabricsnapshot",
            name="created",
            field=models.DateTimeField(auto_now_add=True, null=True),
        ),
        migrations.AlterField(
            model_name="ipfabricsnapshot",
            name="last_updated",
            field=models.DateTimeField(auto_now=True, null=True),
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from netbox.api.fields import ChoiceField
from netbox.api.fields import ContentTypeField
from netbox.api.fields import RelatedObjectCountField
from netbox.api.serializers import NestedGroupModelSerializer
from netbox_branching.api.serializers import BranchSerializer
from rest_framework import serializers

from ipfabric_netbox.choices import IPFabricSourceStatusChoices
from ipfabric_netbox.choices import IPFabricSyncStatusChoices
from ipfabric_netbox.models import IPFabricEndpoint
from ipfabric_netbox.models import IPFabricFilter
from ipfabric_netbox.models import IPFabricFilterExpression
from ipfabric_netbox.models import IPFabricIngestion
from ipfabric_netbox.models import IPFabricIngestionIssue
from ipfabric_netbox.models import IPFabricRelationshipField
from ipfabric_netbox.models import IPFabricRelationshipFieldSourceModels
from ipfabric_netbox.models import IPFabricSnapshot
from ipfabric_netbox.models import IPFabricSource
from ipfabric_netbox.models import IPFabricSupportedSyncModels
from ipfabric_netbox.models import IPFabricSync
from ipfabric_netbox.models import IPFabricTransformField
from ipfabric_netbox.models import IPFabricTransformMap
from ipfabric_netbox.models import IPFabricTransformMapGroup

__all__ = (
    "IPFabricEndpointSerializer",
    "IPFabricFilterSerializer",
    "IPFabricFilterExpressionSerializer",
    "IPFabricIngestionSerializer",
    "IPFabricIngestionIssueSerializer",
    "IPFabricRelationshipFieldSerializer",
    "IPFabricSnapshotSerializer",
    "IPFabricSourceSerializer",
    "IPFabricSyncSerializer",
    "IPFabricTransformFieldSerializer",
    "IPFabricTransformMapSerializer",
    "IPFabricTransformMapGroupSerializer",
)


class IPFabricEndpointSerializer(NestedGroupModelSerializer):
    class Meta:
        model = IPFabricEndpoint
        fields = (
            "id",
            "name",
            "display",
            "description",
            "endpoint",
            "created",
            "last_updated",
        )
        brief_fields = (
            "id",
            "name",
            "display",
            "endpoint",
        )


class IPFabricTransformMapGroupSerializer(NestedGroupModelSerializer):
    transform_maps_count = RelatedObjectCountField("transform_maps")

    class Meta:
        model = IPFabricTransformMapGroup
        fields = (
            "id",
            "name",
            "display",
            "description",
            "transform_maps_count",
            "created",
            "last_updated",
        )
        brief_fields = (
            "id",
            "name",
            "display",
            "description",
        )


class IPFabricTransformMapSerializer(NestedGroupModelSerializer):
    group = IPFabricTransformMapGroupSerializer(
        nested=True, required=False, allow_null=True
    )
    source_endpoint = IPFabricEndpointSerializer(nested=True, required=True)
    target_model = ContentTypeField(
        queryset=ContentType.objects.filter(IPFabricSupportedSyncModels)
    )
    parents = serializers.PrimaryKeyRelatedField(
        queryset=IPFabricTransformMap.objects.all(),
        many=True,
        required=False,
        allow_null=True,
    )

    class Meta:
        model = IPFabricTransformMap
        fields = (
            "id",
            "name",
            "display",
            "group",
            "parents",
            "source_endpoint",
            "target_model",
            "created",
            "last_updated",
        )
        brief_fields = (
            "id",
            "name",
            "display",
            "group",
            "source_endpoint",
            "target_model",
        )


class IPFabricTransformFieldSerializer(NestedGroupModelSerializer):
    transform_map = IPFabricTransformMapSerializer(nested=True)

    class Meta:
        model = IPFabricTransformField
        fields = (
            "id",
            "transform_map",
            "display",
            "source_field",
            "target_field",
            "coalesce",
            "template",
        )


class IPFabricRelationshipFieldSerializer(NestedGroupModelSerializer):
    transform_map = IPFabricTransformMapSerializer(nested=True)
    source_model = ContentTypeField(
        queryset=ContentType.objects.filter(IPFabricRelationshipFieldSourceModels)
    )

    class Meta:
        model = IPFabricRelationshipField
        fields = (
            "id",
            "transform_map",
            "display",
            "source_model",
            "target_field",
            "coalesce",
            "template",
        )


class IPFabricSourceSerializer(NestedGroupModelSerializer):
    status = ChoiceField(choices=IPFabricSourceStatusChoices, read_only=True)
    url = serializers.URLField()

    class Meta:
        model = IPFabricSource
        fields = (
            "id",
            "url",
            "display",
            "name",
            "type",
            "status",
            "last_synced",
            "description",
            "comments",
            "parameters",
            "created",
            "last_updated",
        )
        brief_fields = (
            "display",
            "id",
            "name",
            "status",
            "type",
            "url",
        )


class IPFabricSnapshotSerializer(NestedGroupModelSerializer):
    source = IPFabricSourceSerializer(nested=True, read_only=True)
    data = serializers.JSONField()
    display = serializers.CharField(source="__str__", read_only=True)

    class Meta:
        model = IPFabricSnapshot
        fields = (
            "id",
            "display",
            "name",
            "source",
            "snapshot_id",
            "status",
            "data",
            "date",
            "created",
            "last_updated",
        )
        brief_fields = (
            "display",
            "id",
            "name",
            "source",
            "snapshot_id",
            "status",
            "data",
            "date",
        )


class IPFabricSyncSerializer(NestedGroupModelSerializer):
    status = ChoiceField(choices=IPFabricSyncStatusChoices, read_only=True)
    snapshot_data = IPFabricSnapshotSerializer(nested=True)
    parameters = serializers.JSONField()

    class Meta:
        model = IPFabricSync
        fields = (
            "id",
            "name",
            "display",
            "snapshot_data",
            "status",
            "parameters",
            "auto_merge",
            "last_synced",
            "scheduled",
            "interval",
            "user",
            "filters",
        )
        brief_fields = (
            "auto_merge",
            "id",
            "last_synced",
            "name",
            "display",
            "parameters",
            "status",
        )


class IPFabricFilterSerializer(NestedGroupModelSerializer):
    expressions = serializers.SerializerMethodField(read_only=True)

    def get_expressions(self, obj):
        """Return the related expressions for this filter."""
        return [{"id": expr.id, "name": expr.name} for expr in obj.expressions.all()]

    class Meta:
        model = IPFabricFilter
        fields = (
            "id",
            "name",
            "display",
            "description",
            "endpoints",
            "filter_type",
            "syncs",
            "expressions",
            "created",
            "last_updated",
            "tags",
        )
        brief_fields = (
            "id",
            "name",
            "display",
            "endpoints",
            "filter_type",
            "syncs",
            "expressions",
        )


class IPFabricFilterExpressionSerializer(NestedGroupModelSerializer):
    expression = serializers.JSONField()

    class Meta:
        model = IPFabricFilterExpression
        fields = (
            "id",
            "name",
            "display",
            "description",
            "filters",
            "expression",
            "created",
            "last_updated",
            "tags",
        )
        brief_fields = (
            "id",
            "name",
            "display",
            "filters",
            "expression",
        )


class IPFabricIngestionSerializer(NestedGroupModelSerializer):
    branch = BranchSerializer(read_only=True)
    sync = IPFabricSyncSerializer(nested=True)

    class Meta:
        model = IPFabricIngestion
        fields = (
            "id",
            "name",
            "display",
            "branch",
            "sync",
        )
        brief_fields = (
            "id",
            "name",
            "display",
            "branch",
            "sync",
        )


class IPFabricIngestionIssueSerializer(NestedGroupModelSerializer):
    ingestion = IPFabricIngestionSerializer(nested=True)

    class Meta:
        model = IPFabricIngestionIssue
        fields = (
            "id",
            "ingestion",
            "display",
            "timestamp",
            "model",
            "message",
            "raw_data",
            "coalesce_fields",
            "defaults",
            "exception",
        )
        brief_fields = (
            "exception",
            "id",
            "ingestion",
            "display",
            "message",
            "model",
        )


class EmptySerializer(serializers.Serializer):
    pass



from typing import TYPE_CHECKING

from django.db import migrations
from extras.choices import CustomFieldTypeChoices
from extras.choices import CustomFieldUIEditableChoices
from extras.choices import CustomFieldUIVisibleChoices
from extras.choices import CustomLinkButtonClassChoices


if TYPE_CHECKING:
    from django.apps import apps
    from django.db.backends.base.schema import BaseDatabaseSchemaEditor


def create_custom_field(
    apps: "apps",
    field_name: str,
    label: str,
    models: list,
    object_type=None,
    cf_type: str | None = "type_text",
):
    """Create a single custom field and link it to required models."""
    ObjectType = apps.get_model("core", "ObjectType")

    defaults = {
        "label": label,
        "related_object_type": (
            ObjectType.objects.get_for_model(object_type) if object_type else None
        ),
        "ui_visible": getattr(CustomFieldUIVisibleChoices, "ALWAYS"),
        "ui_editable": getattr(CustomFieldUIEditableChoices, "NO"),
    }

    custom_field, _ = apps.get_model("extras", "CustomField").objects.update_or_create(
        type=getattr(CustomFieldTypeChoices, cf_type.upper()),
        name=field_name,
        defaults=defaults,
    )

    for model in models:
        custom_field.object_types.add(ObjectType.objects.get_for_model(model))


def prepare_custom_fields(apps: "apps", schema_editor: "BaseDatabaseSchemaEditor"):
    """Forward migration to prepare ipfabric_netbox custom fields and links."""
    Device = apps.get_model("dcim", "Device")
    Site = apps.get_model("dcim", "Site")

    create_custom_field(
        apps,
        "ipfabric_source",
        "IP Fabric Source",
        [Device, Site],
        cf_type="type_object",
        object_type=apps.get_model("ipfabric_netbox", "IPFabricSource"),
    )
    create_custom_field(
        apps,
        "ipfabric_branch",
        "IP Fabric Last Sync",
        [Device, Site],
        cf_type="type_object",
        object_type=apps.get_model("ipfabric_netbox", "IPFabricBranch"),
    )
    cl, _ = apps.get_model("extras", "CustomLink").objects.update_or_create(
        defaults={
            "link_text": "{% if object.custom_field_data.ipfabric_source is defined %}{% set SOURCE_ID = object.custom_field_data.ipfabric_source %}{% if SOURCE_ID %}IP Fabric{% endif %}{% endif %}",
            "link_url": '{% if object.custom_field_data.ipfabric_source is defined %}{% set SOURCE_ID = object.custom_field_data.ipfabric_source %}{% if SOURCE_ID %}{% set BASE_URL = object.custom_fields.filter(related_object_type__model="ipfabricsource").first().related_object_type.model_class().objects.get(pk=SOURCE_ID).url %}{{ BASE_URL }}/inventory/devices?options={"filters":{"sn": ["like","{{ object.serial }}"]}}{% endif %}{%endif%}',
            "new_window": True,
            "button_class": CustomLinkButtonClassChoices.BLUE,
        },
        name="ipfabric",
    )
    cl.object_types.add(
        apps.get_model("core", "ObjectType").objects.get_for_model(Device)
    )


def cleanup_custom_fields(apps: "apps", schema_editor: "BaseDatabaseSchemaEditor"):
    """Reverse migration to prepare ipfabric_netbox custom fields and links."""
    for custom_field_name in ["ipfabric_source", "ipfabric_branch"]:
        custom_field = apps.get_model("extras", "CustomField").objects.get(
            name=custom_field_name
        )
        for model in custom_field.object_types.all()[:]:
            custom_field.object_types.remove(model)
        custom_field.delete()


class Migration(migrations.Migration):
    dependencies = [
        ("dcim", "0191_module_bay_rebuild"),
        ("extras", "0121_customfield_related_object_filter"),
        (
            "ipfabric_netbox",
            "0006_alter_ipfabrictransformmap_target_model",
        ),
    ]
    operations = [
        migrations.RunPython(prepare_custom_fields, cleanup_custom_fields),
    ]
